"""
Predefined useful physical constants

Note that all of these names can be imported from the top-level unyt namespace.
For example::

    >>> from unyt import c, G, kb

"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------


from math import pi as _pi

from unyt._physical_ratios import (
    mass_electron_kg as _mass_electron_kg,
    amu_kg as _amu_kg,
    mass_hydrogen_kg as _mass_hydrogen_kg,
    speed_of_light_m_per_s as _speed_of_light_m_per_s,
    boltzmann_constant_J_per_K as _boltzmann_constant_J_per_K,
    mass_sun_kg as _mass_sun_kg,
    mass_jupiter_kg as _mass_jupiter_kg,
    mass_mercury_kg as _mass_mercury_kg,
    mass_venus_kg as _mass_venus_kg,
    mass_earth_kg as _mass_earth_kg,
    mass_mars_kg as _mass_mars_kg,
    mass_saturn_kg as _mass_saturn_kg,
    mass_uranus_kg as _mass_uranus_kg,
    mass_neptune_kg as _mass_neptune_kg,
    planck_mass_kg as _planck_mass_kg,
    planck_length_m as _planck_length_m,
    planck_time_s as _planck_time_s,
    planck_energy_J as _planck_energy_J,
    planck_charge_C as _planck_charge_C,
    planck_temperature_K as _planck_temperature_K,
    standard_gravity_m_per_s2 as _standard_gravity_m_per_s2,
    newton_mks as _newton_mks,
    planck_mks as _planck_mks,
    eps_0 as _eps_0,
    mu_0 as _mu_0
)
from unyt.array import unyt_quantity

#: mass of the electron
mass_electron_mks = mass_electron = me = \
    unyt_quantity(_mass_electron_kg, 'kg')

#: atomic mass unit
amu_mks = amu = unyt_quantity(_amu_kg, 'kg')

#: Avogadro's number
Na = avogadros_number = unyt_quantity(6.02214085774*10**23, 'mol**-1')

#: Mass of hydrogen
mp = mh = mass_hydrogen = mass_hydrogen_mks = \
    unyt_quantity(_mass_hydrogen_kg, 'kg')

# Velocities
#: speed of light
c = clight = speed_of_light = speed_of_light_mks = \
    unyt_quantity(_speed_of_light_m_per_s, 'm/s')

# Cross Sections
#: Thompson cross section (8*pi/3 (alpha*hbar*c/(2*pi))**2)
sigma_thompson = thompson_cross_section = cross_section_thompson = \
    cross_section_thompson_mks = unyt_quantity(6.65245854533e-29, 'm**2')

# Charge
#: Charge of the proton
qp = elementary_charge = proton_charge = charge_proton = charge_proton_mks = \
    unyt_quantity(1.6021766208e-19, 'C')
#: Charge of the electron
electron_charge = charge_electron = charge_electron_mks = -qp

qp_cgs = elementary_charge_cgs = proton_charge_cgs = charge_proton_cgs = \
    unyt_quantity(4.8032056e-10, 'esu')
electron_charge_cgs = charge_electron_cgs = -charge_proton_cgs

# Physical Constants
#: Boltzmann constant
kb = kboltz = boltzmann_constant = boltzmann_constant_mks = unyt_quantity(
    _boltzmann_constant_J_per_K, 'J/K')

#: Gravitational constant
G = newtons_constant = gravitational_constant = gravitational_constant_mks = \
    unyt_quantity(_newton_mks, 'm**3/kg/s**2')

#: Planck constant
hmks = planck_constant = planck_constant_mks = \
    unyt_quantity(_planck_mks, 'J*s')

#: Reduced Planck constant
hbar = reduced_planck_constant = 0.5*hmks/_pi

#: Stefan-Boltzmann constant
stefan_boltzmann_constant = stefan_boltzmann_constant_mks = unyt_quantity(
    5.670373e-8, 'W/m**2/K**4')

#: Current CMB temperature
CMD_temperature = Tcmb = unyt_quantity(2.726, 'K')

# Solar System
#: Mass of the sun
msun = solar_mass = mass_sun = mass_sun_mks = unyt_quantity(
    _mass_sun_kg, 'kg')

#: Mass of Jupiter
Mjup = mjup = jupiter_mass = mass_jupiter = mass_jupiter_mks = unyt_quantity(
    _mass_jupiter_kg, 'kg')

#: Mass of Mercury
mercury_mass = mass_mercury = mass_mercury_mks = unyt_quantity(
    _mass_mercury_kg, 'kg')

#: Mass of Venus
venus_mass = mass_venus = mass_venus_mks = unyt_quantity(
    _mass_venus_kg, 'kg')

#: Mass of Earth
Mearth = mearth = earth_mass = mass_earth = mass_earth_mks = unyt_quantity(
    _mass_earth_kg, 'kg')

#: Mass of Mars
mars_mass = mass_mars = mass_mars_mks = unyt_quantity(_mass_mars_kg, 'kg')

#: Mass of Saturn
saturn_mass = mass_saturn = mass_saturn_mks = unyt_quantity(
    _mass_saturn_kg, 'kg')

#: Mass of Uranus
uranus_mass = mass_uranus = mass_uranus_mks = unyt_quantity(
    _mass_uranus_kg, 'kg')

#: Mass of Neptune
neptune_mass = mass_neptune = mass_neptune_mks = unyt_quantity(
    _mass_neptune_kg, 'kg')

# Planck units
#: Planck mass
m_pl = planck_mass = unyt_quantity(_planck_mass_kg, "g")
#: Planck length
l_pl = planck_length = unyt_quantity(_planck_length_m, "cm")
#: Planck time
t_pl = planck_time = unyt_quantity(_planck_time_s, "s")
#: Planck energy
E_pl = planck_energy = unyt_quantity(_planck_energy_J, "erg")
#: Planck charge
q_pl = planck_charge = unyt_quantity(_planck_charge_C, "C")
#: Planck tempearture
T_pl = planck_temperature = unyt_quantity(_planck_temperature_K, "K")

# MKS E&M constants
#: Permeability of Free Space
mu_0 = unyt_quantity(_mu_0, "N/A**2")
#: Permittivity of Free Space
eps_0 = unyt_quantity(_eps_0, "C**2/N/m**2")

# Misc
#: Standard gravitational acceleration
standard_gravity = standard_gravity_mks = unyt_quantity(
    _standard_gravity_m_per_s2, "m/s**2")



"""
The default unit symbol lookup table.


"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from unyt import dimensions
from unyt._physical_ratios import (
    m_per_pc,
    m_per_ly,
    m_per_au,
    m_per_rsun,
    m_per_inch,
    m_per_ft,
    watt_per_horsepower,
    mass_sun_kg,
    sec_per_year,
    sec_per_day,
    sec_per_hr,
    sec_per_min,
    temp_sun_kelvin,
    luminosity_sun_watts,
    metallicity_sun,
    J_per_eV,
    amu_kg,
    amu_grams,
    mass_electron_kg,
    m_per_ang,
    jansky_mks,
    mass_jupiter_kg,
    mass_earth_kg,
    kelvin_per_rankine,
    speed_of_light_m_per_s,
    planck_length_m,
    planck_charge_C,
    planck_energy_J,
    planck_mass_kg,
    planck_temperature_K,
    planck_time_s,
    mass_hydrogen_kg,
    kg_per_pound,
    standard_gravity_m_per_s2,
    pascal_per_atm,
    newton_mks,
    m_per_rearth,
    m_per_rjup
)
import numpy as np

# Lookup a unit symbol with the symbol string, and provide a tuple with the
# conversion factor to cgs and dimensionality.

default_unit_symbol_lut = {
    # base
    "m": (1.0, dimensions.length, 0.0, r"\rm{m}", True),
    "g":  (1.0e-3, dimensions.mass, 0.0, r"\rm{g}", True),
    "s":  (1.0, dimensions.time, 0.0, r"\rm{s}", True),
    "K":  (1.0, dimensions.temperature, 0.0, r"\rm{K}", True),
    "radian": (1.0, dimensions.angle, 0.0, r"\rm{radian}", True),
    "A": (1.0, dimensions.current_mks, 0.0, r"\rm{A}", True),
    "cd": (1.0, dimensions.luminous_intensity, 0.0, r"\rm{cd}", True),
    "mol": (1.0 / amu_grams, dimensions.dimensionless, 0.0, r"\rm{mol}", True),

    # some cgs
    "dyne": (1.0e-5, dimensions.force, 0.0, r"\rm{dyn}", True),
    "erg":  (1.0e-7, dimensions.energy, 0.0, r"\rm{erg}", True),
    "Ba": (0.1, dimensions.pressure, 0.0, r"\rm{Ba}", True),
    "esu":  (1.0e-3**1.5, dimensions.charge_cgs, 0.0, r"\rm{esu}", True),
    "gauss": (0.1**0.5, dimensions.magnetic_field_cgs, 0.0, r"\rm{G}", True),
    "degC": (1.0, dimensions.temperature, -273.15, r"^\circ\rm{C}", True),
    "statA": (1.0e-3**1.5, dimensions.current_cgs, 0.0, r"\rm{statA}", True),
    "statV": (0.1*1.0e-3**0.5, dimensions.electric_potential_cgs,
              0.0, r"\rm{statV}", True),
    "statohm": (100.0, dimensions.resistance_cgs, 0.0, r"\rm{statohm}", True),
    "Mx": (1.0e-3**1.5, dimensions.magnetic_flux_cgs, 0.0, r"\rm{Mx}", True),

    # some SI
    "J": (1.0, dimensions.energy, 0.0, r"\rm{J}", True),
    "W": (1.0, dimensions.power, 0.0, r"\rm{W}", True),
    "Hz": (1.0, dimensions.rate, 0.0, r"\rm{Hz}", True),
    "N": (1.0, dimensions.force, 0.0, r"\rm{N}", True),
    "C": (1.0, dimensions.charge_mks, 0.0, r"\rm{C}", True),
    "T": (1.0, dimensions.magnetic_field_mks, 0.0, r"\rm{T}", True),
    "Pa": (1.0, dimensions.pressure, 0.0, r"\rm{Pa}", True),
    "V": (1.0, dimensions.electric_potential_mks, 0.0, r"\rm{V}", True),
    "ohm": (1.0, dimensions.resistance_mks, 0.0, r"\Omega", True),
    "Wb": (1.0, dimensions.magnetic_flux_mks, 0.0, r"\rm{Wb}", True),
    "lm": (1.0, dimensions.luminous_flux, 0.0, r"\rm{lm}", True),
    "lx": (1.0, dimensions.luminous_flux/dimensions.area, 0.0, r"\rm{lx}",
           True),

    # Imperial and other non-metric units
    "inch": (m_per_inch, dimensions.length, 0.0, r"\rm{in}", False),
    "ft": (m_per_ft, dimensions.length, 0.0, r"\rm{ft}", False),
    "yd": (0.9144, dimensions.length, 0.0, r"\rm{yd}", False),
    "mile": (1609.344, dimensions.length, 0.0, r"\rm{mile}", False),
    "fur": (m_per_ft*660.0, dimensions.length, 0.0, r"\rm{fur}", False),
    "degF": (kelvin_per_rankine, dimensions.temperature, -459.67,
             "^\circ\rm{F}", False),
    "R": (kelvin_per_rankine, dimensions.temperature, 0.0, r"^\circ\rm{R}",
          False),
    "lbf": (kg_per_pound*standard_gravity_m_per_s2,
            dimensions.force, 0.0, r"\rm{lbf}", False),
    "lb": (kg_per_pound, dimensions.mass, 0.0, r"\rm{lb}", False),
    "lbm": (kg_per_pound, dimensions.mass, 0.0, r"\rm{lbm}", False),
    "atm": (pascal_per_atm, dimensions.pressure, 0.0, r"\rm{atm}", False),
    "hp": (watt_per_horsepower, dimensions.power, 0.0, r"\rm{hp}", False),
    "oz": (kg_per_pound/16.0, dimensions.mass, 0.0, r"\rm{oz}", False),
    "ton": (kg_per_pound*2000.0, dimensions.mass, 0.0, r"\rm{ton}", False),
    "slug": (kg_per_pound*standard_gravity_m_per_s2/m_per_ft,
             dimensions.mass, 0.0, r"\rm{slug}", False),
    "cal": (4.184, dimensions.energy, 0.0, r"\rm{cal}", True),
    "BTU": (1055.0559, dimensions.energy, 0.0, r"\rm{BTU}", False),
    "psi": (kg_per_pound*standard_gravity_m_per_s2/m_per_inch**2,
            dimensions.pressure, 0.0, r"\rm{psi}", False),
    "smoot": (1.7018, dimensions.length, 0.0, r"\rm{smoot}", False),

    # dimensionless stuff
    "h": (1.0, dimensions.dimensionless, 0.0, r"h", False),
    "dimensionless": (1.0, dimensions.dimensionless, 0.0, r"", False),

    # times
    "min": (sec_per_min, dimensions.time, 0.0, r"\rm{min}", False),
    "hr":  (sec_per_hr, dimensions.time, 0.0, r"\rm{hr}", False),
    "day": (sec_per_day, dimensions.time, 0.0, r"\rm{d}", False),
    "d": (sec_per_day, dimensions.time, 0.0, r"\rm{d}", False),
    "yr":  (sec_per_year, dimensions.time, 0.0, r"\rm{yr}", True),

    # Velocities
    "c": (speed_of_light_m_per_s, dimensions.velocity, 0.0, r"\rm{c}", False),

    # Solar units
    "Msun": (mass_sun_kg, dimensions.mass, 0.0, r"M_\odot", False),
    "msun": (mass_sun_kg, dimensions.mass, 0.0, r"M_\odot", False),
    "Rsun": (m_per_rsun, dimensions.length, 0.0, r"R_\odot", False),
    "rsun": (m_per_rsun, dimensions.length, 0.0, r"R_\odot", False),
    "R_sun": (m_per_rsun, dimensions.length, 0.0, r"R_\odot", False),
    "r_sun": (m_per_rsun, dimensions.length, 0.0, r"R_\odot", False),
    "Lsun": (luminosity_sun_watts, dimensions.power, 0.0, r"L_\odot", False),
    "Tsun": (temp_sun_kelvin, dimensions.temperature, 0.0, r"T_\odot", False),
    "Zsun": (metallicity_sun, dimensions.dimensionless, 0.0, r"Z_\odot",
             False),
    "Mjup": (mass_jupiter_kg, dimensions.mass, 0.0, r"M_{\rm{Jup}}", False),
    "Mearth": (mass_earth_kg, dimensions.mass, 0.0, r"M_\oplus", False),
    "R_jup": (m_per_rjup, dimensions.length, 0.0, r"R_\mathrm{Jup}", False),
    "r_jup": (m_per_rjup, dimensions.length, 0.0, r"R_\mathrm{Jup}", False),
    "R_earth": (m_per_rearth, dimensions.length, 0.0, r"R_\oplus", False),
    "r_earth": (m_per_rearth, dimensions.length, 0.0, r"R_\oplus", False),

    # astro distances
    "AU": (m_per_au, dimensions.length, 0.0, r"\rm{AU}", False),
    "au": (m_per_au, dimensions.length, 0.0, r"\rm{AU}", False),
    "ly": (m_per_ly, dimensions.length, 0.0, r"\rm{ly}", False),
    "pc": (m_per_pc, dimensions.length, 0.0, r"\rm{pc}", True),

    # angles
    "degree": (np.pi/180., dimensions.angle, 0.0, r"\rm{deg}", False),
    "arcmin": (np.pi/10800., dimensions.angle, 0.0,
               r"\rm{arcmin}", False),  # arcminutes
    "arcsec": (np.pi/648000., dimensions.angle, 0.0,
               r"\rm{arcsec}", False),  # arcseconds
    "mas": (np.pi/648000000., dimensions.angle, 0.0,
            r"\rm{mas}", False),  # milliarcseconds
    "hourangle": (np.pi/12., dimensions.angle, 0.0, r"\rm{HA}", False),
    "steradian": (1.0, dimensions.solid_angle, 0.0, r"\rm{sr}", False),
    "lat": (-np.pi/180.0, dimensions.angle, 90.0, r"\rm{Latitude}", False),
    "lon": (np.pi/180.0, dimensions.angle, -180.0, r"\rm{Longitude}", False),

    # misc
    "eV": (J_per_eV, dimensions.energy, 0.0, r"\rm{eV}", True),
    "amu": (amu_kg, dimensions.mass, 0.0, r"\rm{amu}", False),
    "angstrom": (m_per_ang, dimensions.length, 0.0, r"\AA", False),
    "Jy": (jansky_mks, dimensions.specific_flux, 0.0, r"\rm{Jy}", True),
    "counts": (1.0, dimensions.dimensionless, 0.0, r"\rm{counts}", False),
    "photons": (1.0, dimensions.dimensionless, 0.0, r"\rm{photons}", False),
    "me": (mass_electron_kg, dimensions.mass, 0.0, r"m_e", False),
    "mp": (mass_hydrogen_kg, dimensions.mass, 0.0, r"m_p", False),
    'Sv': (1.0, dimensions.specific_energy, 0.0, r"\rm{Sv}", True),
    "rayleigh": (2.5e9/np.pi, dimensions.count_intensity, 0.0, r"\rm{R}",
                 False),
    "lambert": (1.0e4/np.pi, dimensions.luminance, 0.0, r"\rm{L}", False),
    "nt": (1.0, dimensions.luminance, 0.0, r"\rm{nt}", False),

    # for AstroPy compatibility
    "solMass": (mass_sun_kg, dimensions.mass, 0.0, r"M_\odot", False),
    "solRad": (m_per_rsun, dimensions.length, 0.0, r"R_\odot", False),
    "solLum": (luminosity_sun_watts, dimensions.power, 0.0, r"L_\odot", False),
    "dyn": (1.0e-5, dimensions.force, 0.0, r"\rm{dyn}", False),
    "sr": (1.0, dimensions.solid_angle, 0.0, r"\rm{sr}", False),
    "rad": (1.0, dimensions.angle, 0.0, r"\rm{rad}", False),
    "deg": (np.pi/180., dimensions.angle, 0.0, r"\rm{deg}", False),
    "Fr":  (1.0e-3**1.5, dimensions.charge_cgs, 0.0, r"\rm{Fr}", False),
    "G": (0.1**0.5, dimensions.magnetic_field_cgs, 0.0, r"\rm{G}", True),
    "Angstrom": (m_per_ang, dimensions.length, 0.0, r"\AA", False),
    "statC": (1.0e-3**1.5, dimensions.charge_cgs, 0.0, r"\rm{statC}", True),

    # Planck units
    "m_pl": (planck_mass_kg, dimensions.mass, 0.0, r"m_{\rm{P}}", False),
    "l_pl": (planck_length_m, dimensions.length, 0.0, r"\ell_\rm{P}", False),
    "t_pl": (planck_time_s, dimensions.time, 0.0, r"t_{\rm{P}}", False),
    "T_pl": (planck_temperature_K, dimensions.temperature, 0.0, r"T_{\rm{P}}",
             False),
    "q_pl": (planck_charge_C, dimensions.charge_mks, 0.0, r"q_{\rm{P}}",
             False),
    "E_pl": (planck_energy_J, dimensions.energy, 0.0, r"E_{\rm{P}}", False),

    # Geometrized units
    "m_geom": (mass_sun_kg, dimensions.mass, 0.0, r"M_\odot", False),
    "l_geom": (newton_mks*mass_sun_kg/speed_of_light_m_per_s**2,
               dimensions.length, 0.0, r"M_\odot", False),
    "t_geom": (newton_mks*mass_sun_kg/speed_of_light_m_per_s**3,
               dimensions.time, 0.0, r"M_\odot", False),
}

# This dictionary formatting from magnitude package, credit to Juan Reyero.
unit_prefixes = {
    'Y': 1e24,   # yotta
    'Z': 1e21,   # zetta
    'E': 1e18,   # exa
    'P': 1e15,   # peta
    'T': 1e12,   # tera
    'G': 1e9,    # giga
    'M': 1e6,    # mega
    'k': 1e3,    # kilo
    'h': 1e2,    # hecto
    'da': 1e1,   # deca
    'd': 1e-1,   # deci
    'c': 1e-2,   # centi
    'm': 1e-3,   # mili
    'u': 1e-6,   # micro
    'n': 1e-9,   # nano
    'p': 1e-12,  # pico
    'f': 1e-15,  # femto
    'a': 1e-18,  # atto
    'z': 1e-21,  # zepto
    'y': 1e-24,  # yocto
}

latex_prefixes = {
    "u": r"\mu",
    }

default_base_units = {
    dimensions.mass: 'kg',
    dimensions.length: 'm',
    dimensions.time: 's',
    dimensions.temperature: 'K',
    dimensions.angle: 'radian',
    dimensions.current_mks: 'A',
    dimensions.luminous_intensity: 'cd',
}



"""
Test symbolic unit handling.




"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------


import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_allclose,
    assert_array_almost_equal_nulp,
    assert_equal
)
import operator
import pytest
from six.moves import cPickle as pickle
from sympy import Symbol
from unyt._testing import assert_allclose_units
from unyt.unit_registry import UnitRegistry
from unyt import electrostatic_unit, elementary_charge_cgs
from unyt.dimensions import (
    mass,
    length,
    time,
    temperature,
    energy,
    magnetic_field_cgs,
    magnetic_field_mks,
    power,
    rate
)
from unyt.exceptions import (
    InvalidUnitOperation,
    UnitsNotReducible,
    UnitConversionError,
)
from unyt.unit_object import (
    default_unit_registry,
    Unit,
    UnitParseError,
)
from unyt.unit_systems import UnitSystem
from unyt._unit_lookup_table import (
    default_unit_symbol_lut,
    unit_prefixes
)
import unyt.unit_symbols as unit_symbols
from unyt._physical_ratios import (
    m_per_pc,
    sec_per_year,
    m_per_km,
    m_per_mpc,
    mass_sun_kg
)


def test_no_conflicting_symbols():
    """
    Check unit symbol definitions for conflicts.

    """
    full_set = set(default_unit_symbol_lut.keys())

    # go through all possible prefix combos
    for symbol in default_unit_symbol_lut.keys():
        if default_unit_symbol_lut[symbol][4]:
            keys = unit_prefixes.keys()
        else:
            keys = [symbol]
        for prefix in keys:
            new_symbol = "%s%s" % (prefix, symbol)

            # test if we have seen this symbol
            assert new_symbol not in full_set, ("Duplicate symbol: %s" %
                                                new_symbol)

            full_set.add(new_symbol)


def test_dimensionless():
    """
    Create dimensionless unit and check attributes.

    """
    u1 = Unit()

    assert u1.is_dimensionless
    assert u1.expr == 1
    assert u1.base_value == 1
    assert u1.dimensions == 1
    assert u1 != 'hello!'
    assert (u1 == 'hello') is False

    u2 = Unit("")

    assert u2.is_dimensionless
    assert u2.expr == 1
    assert u2.base_value == 1
    assert u2.dimensions == 1

    assert_equal(u1.latex_repr, '')
    assert_equal(u2.latex_repr, '')


def test_create_from_string():
    """
    Create units with strings and check attributes.

    """

    u1 = Unit("kg * m**2 * s**-2")
    assert u1.dimensions == energy
    assert u1.base_value == 1.0

    # make sure order doesn't matter
    u2 = Unit("m**2 * s**-2 * kg")
    assert u2.dimensions == energy
    assert u2.base_value == 1.0

    # Test rationals
    u3 = Unit("kg**0.5 * m**-0.5 * s**-1")
    assert u3.dimensions == magnetic_field_cgs
    assert u3.base_value == 1.0

    # sqrt functions
    u4 = Unit("sqrt(kg)/sqrt(m)/s")
    assert u4.dimensions == magnetic_field_cgs
    assert u4.base_value == 1.0

    # commutative sqrt function
    u5 = Unit("sqrt(kg/m)/s")
    assert u5.dimensions == magnetic_field_cgs
    assert u5.base_value == 1.0

    # nonzero CGS conversion factor
    u6 = Unit("Msun/pc**3")
    assert u6.dimensions == mass/length**3
    assert_array_almost_equal_nulp(np.array([u6.base_value]),
                                   np.array([mass_sun_kg/m_per_pc**3]))

    with pytest.raises(UnitParseError):
        Unit('m**m')
    with pytest.raises(UnitParseError):
        Unit('m**g')
    with pytest.raises(UnitParseError):
        Unit('m+g')
    with pytest.raises(UnitParseError):
        Unit('m-g')
    with pytest.raises(UnitParseError):
        Unit('hello!')

    cm = Unit('cm')
    data = 1*cm

    assert Unit(data) == cm


def test_create_from_expr():
    """
    Create units from sympy Exprs and check attributes.

    """
    pc_mks = m_per_pc
    yr_mks = sec_per_year

    # Symbol expr
    s1 = Symbol("pc", positive=True)
    s2 = Symbol("yr", positive=True)
    # Mul expr
    s3 = s1 * s2
    # Pow expr
    s4 = s1**2 * s2**(-1)

    u1 = Unit(s1)
    u2 = Unit(s2)
    u3 = Unit(s3)
    u4 = Unit(s4)

    assert u1.expr == s1
    assert u2.expr == s2
    assert u3.expr == s3
    assert u4.expr == s4

    assert_allclose_units(u1.base_value, pc_mks, 1e-12)
    assert_allclose_units(u2.base_value, yr_mks, 1e-12)
    assert_allclose_units(u3.base_value, pc_mks * yr_mks, 1e-12)
    assert_allclose_units(u4.base_value, pc_mks**2 / yr_mks, 1e-12)

    assert u1.dimensions == length
    assert u2.dimensions == time
    assert u3.dimensions == length * time
    assert u4.dimensions == length**2 / time


def test_create_with_duplicate_dimensions():
    """
    Create units with overlapping dimensions. Ex: km/Mpc.

    """

    u1 = Unit("J * s**-1")
    u2 = Unit("km/s/Mpc")
    km_mks = m_per_km
    Mpc_mks = m_per_mpc

    assert u1.base_value == 1
    assert u1.dimensions == power

    assert_allclose_units(u2.base_value, km_mks / Mpc_mks, 1e-12)
    assert u2.dimensions == rate


def test_create_new_symbol():
    """
    Create unit with unknown symbol.

    """
    u1 = Unit("abc", base_value=42, dimensions=(mass/time))

    assert u1.expr == Symbol("abc", positive=True)
    assert u1.base_value == 42
    assert u1.dimensions == mass / time

    u1 = Unit("abc", base_value=42, dimensions=length**3)

    assert u1.expr == Symbol("abc", positive=True)
    assert u1.base_value == 42
    assert u1.dimensions == length**3

    u1 = Unit("abc", base_value=42, dimensions=length*(mass*length))

    assert u1.expr == Symbol("abc", positive=True)
    assert u1.base_value == 42
    assert u1.dimensions == length**2*mass

    with pytest.raises(UnitParseError):
        Unit('abc', base_value=42, dimensions=length**length)
    with pytest.raises(UnitParseError):
        Unit('abc', base_value=42, dimensions=length**(length*length))
    with pytest.raises(UnitParseError):
        Unit('abc', base_value=42, dimensions=length-mass)
    with pytest.raises(UnitParseError):
        Unit('abc', base_value=42, dimensions=length+mass)


def test_create_fail_on_unknown_symbol():
    """
    Fail to create unit with unknown symbol, without base_value and dimensions.

    """
    with pytest.raises(UnitParseError):
        Unit(Symbol("jigawatts"))


def test_create_fail_on_bad_symbol_type():
    """
    Fail to create unit with bad symbol type.

    """
    with pytest.raises(UnitParseError):
        Unit([1])  # something other than Expr and str


def test_create_fail_on_bad_dimensions_type():
    """
    Fail to create unit with bad dimensions type.

    """
    with pytest.raises(UnitParseError):
        Unit("a", base_value=1, dimensions="(mass)")


def test_create_fail_on_dimensions_content():
    """
    Fail to create unit with bad dimensions expr.

    """
    a = Symbol("a")
    with pytest.raises(UnitParseError):
        Unit("a", base_value=1, dimensions=a)


def test_create_fail_on_base_value_type():
    """
    Fail to create unit with bad base_value type.

    """
    with pytest.raises(UnitParseError):
        Unit("a", base_value="a", dimensions=(mass/time))


def test_string_representation():
    """
    Check unit string representation.

    """
    pc = Unit("pc")
    Myr = Unit("Myr")
    speed = pc / Myr
    dimensionless = Unit()

    assert str(pc) == "pc"
    assert str(Myr) == "Myr"
    assert str(speed) == "pc/Myr"
    assert repr(speed) == "pc/Myr"
    assert str(dimensionless) == "dimensionless"
    assert repr(dimensionless) == "(dimensionless)"


def test_multiplication():
    """
    Multiply two units.

    """
    msun_mks = mass_sun_kg
    pc_mks = m_per_pc

    # Create symbols
    msun_sym = Symbol("Msun", positive=True)
    pc_sym = Symbol("pc", positive=True)
    s_sym = Symbol("s", positive=True)

    # Create units
    u1 = Unit("Msun")
    u2 = Unit("pc")

    # Mul operation
    u3 = u1 * u2

    assert u3.expr == msun_sym * pc_sym
    assert_allclose_units(u3.base_value, msun_mks * pc_mks, 1e-12)
    assert u3.dimensions == mass * length

    # Pow and Mul operations
    u4 = Unit("pc**2")
    u5 = Unit("Msun * s")

    u6 = u4 * u5

    assert u6.expr == pc_sym**2 * msun_sym * s_sym
    assert_allclose_units(u6.base_value, pc_mks**2 * msun_mks, 1e-12)
    assert u6.dimensions == length**2 * mass * time


def test_division():
    """
    Divide two units.

    """
    pc_mks = m_per_pc
    km_mks = m_per_km

    # Create symbols
    pc_sym = Symbol("pc", positive=True)
    km_sym = Symbol("km", positive=True)
    s_sym = Symbol("s", positive=True)

    # Create units
    u1 = Unit("pc")
    u2 = Unit("km * s")

    u3 = u1 / u2

    assert u3.expr == pc_sym / (km_sym * s_sym)
    assert_allclose_units(u3.base_value, pc_mks / km_mks, 1e-12)
    assert u3.dimensions == 1 / time


def test_power():
    """
    Take units to some power.

    """
    from sympy import nsimplify

    pc_mks = m_per_pc
    mK_mks = 1e-3
    u1_dims = mass * length**2 * time**-3 * temperature**4
    u1 = Unit("kg * pc**2 * s**-3 * mK**4")

    u2 = u1**2

    assert u2.dimensions == u1_dims**2
    assert_allclose_units(u2.base_value, (pc_mks**2 * mK_mks**4)**2, 1e-12)

    u3 = u1**(-1.0/3)

    assert u3.dimensions == nsimplify(u1_dims**(-1.0/3))
    assert_allclose_units(u3.base_value, (pc_mks**2 * mK_mks**4)**(-1.0/3),
                          1e-12)


def test_equality():
    """
    Check unit equality with different symbols, but same dimensions and
    base_value.

    """
    u1 = Unit("km * s**-1")
    u2 = Unit("m * ms**-1")

    assert u1 == u2


def test_invalid_operations():
    u1 = Unit('cm')
    u2 = Unit('m')

    with pytest.raises(InvalidUnitOperation):
        u1+u2
    with pytest.raises(InvalidUnitOperation):
        u1 += u2
    with pytest.raises(InvalidUnitOperation):
        1+u1
    with pytest.raises(InvalidUnitOperation):
        u1+1
    with pytest.raises(InvalidUnitOperation):
        u1-u2
    with pytest.raises(InvalidUnitOperation):
        u1 -= u2
    with pytest.raises(InvalidUnitOperation):
        1-u1
    with pytest.raises(InvalidUnitOperation):
        u1-1
    with pytest.raises(InvalidUnitOperation):
        u1 *= u2
    with pytest.raises(InvalidUnitOperation):
        u1*'hello!'
    with pytest.raises(InvalidUnitOperation):
        u1 /= u2
    with pytest.raises(InvalidUnitOperation):
        u1/'hello!'


def test_base_equivalent():
    """
    Check base equivalent of a unit.

    """
    Msun_mks = mass_sun_kg
    Mpc_mks = m_per_mpc

    u1 = Unit("Msun * Mpc**-3")
    u2 = Unit("kg * m**-3")
    u3 = u1.get_base_equivalent()

    assert u2.expr == u3.expr
    assert u2 == u3

    assert_allclose_units(u1.base_value, Msun_mks / Mpc_mks**3, 1e-12)
    assert u2.base_value == 1
    assert u3.base_value == 1

    mass_density = mass / length**3

    assert u1.dimensions == mass_density
    assert u2.dimensions == mass_density
    assert u3.dimensions == mass_density

    assert_allclose_units(u1.get_conversion_factor(u3)[0],
                          Msun_mks / Mpc_mks**3, 1e-12)

    with pytest.raises(UnitConversionError):
        u1.get_conversion_factor(Unit('m'))

    with pytest.raises(UnitConversionError):
        u1.get_conversion_factor(Unit('degF'))


def test_temperature_offsets():
    u1 = Unit('degC')
    u2 = Unit('degF')

    with pytest.raises(InvalidUnitOperation):
        operator.mul(u1, u2)
    with pytest.raises(InvalidUnitOperation):
        operator.truediv(u1, u2)


def test_latex_repr():
    registry = UnitRegistry()

    # create a fake comoving unit
    registry.add('pccm', registry.lut['pc'][0]/(1+2), length,
                 "\\rm{pc}/(1+z)", prefixable=True)

    test_unit = Unit('Mpccm', registry=registry)
    assert_almost_equal(test_unit.base_value, m_per_mpc/3)
    assert_equal(test_unit.latex_repr, r'\rm{Mpc}/(1+z)')

    test_unit = Unit('cm**-3', base_value=1.0, registry=registry)
    assert_equal(test_unit.latex_repr, '\\frac{1}{\\rm{cm}^{3}}')

    test_unit = Unit('m_geom/l_geom**3')
    assert_equal(test_unit.latex_repr, '\\frac{1}{M_\\odot^{2}}')

    test_unit = Unit('1e9*cm')
    assert_equal(test_unit.latex_repr, '1.0 \\times 10^{9}\\ \\rm{cm}')

    test_unit = Unit('1.0*cm')
    assert_equal(test_unit.latex_repr, '\\rm{cm}')


def test_latitude_longitude():
    lat = unit_symbols.lat
    lon = unit_symbols.lon
    deg = unit_symbols.deg
    assert_equal(lat.units.base_offset, 90.0)
    assert_equal((deg*90.0).in_units("lat").value, 0.0)
    assert_equal((deg*180).in_units("lat").value, -90.0)
    assert_equal((lat*0.0).in_units("deg"), deg*90.0)
    assert_equal((lat*-90).in_units("deg"), deg*180)

    assert_equal(lon.units.base_offset, -180.0)
    assert_equal((deg*0.0).in_units("lon").value, -180.0)
    assert_equal((deg*90.0).in_units("lon").value, -90.0)
    assert_equal((deg*180).in_units("lon").value, 0.0)
    assert_equal((deg*360).in_units("lon").value, 180.0)

    assert_equal((lon*-180.0).in_units("deg"), deg*0.0)
    assert_equal((lon*-90.0).in_units("deg"), deg*90.0)
    assert_equal((lon*0.0).in_units("deg"), deg*180.0)
    assert_equal((lon*180.0).in_units("deg"), deg*360)


def test_registry_json():
    reg = UnitRegistry()
    json_reg = reg.to_json()
    unserialized_reg = UnitRegistry.from_json(json_reg)

    assert_equal(reg.lut, unserialized_reg.lut)


def test_creation_from_ytarray():
    u1 = Unit(electrostatic_unit)
    assert_equal(str(u1), 'esu')
    assert_equal(u1, Unit('esu'))
    assert_equal(u1, electrostatic_unit.units)

    u2 = Unit(elementary_charge_cgs)
    assert_equal(str(u2), '4.8032056e-10*esu')
    assert_equal(u2, Unit('4.8032056e-10*esu'))
    assert_equal(u1, elementary_charge_cgs.units)

    assert_allclose((u1/u2).base_value,
                    electrostatic_unit/elementary_charge_cgs)

    with pytest.raises(UnitParseError):
        Unit([1, 2, 3]*elementary_charge_cgs)


def test_list_same_dimensions():
    from unyt import m
    reg = default_unit_registry
    for equiv in reg.list_same_dimensions(m):
        assert Unit(equiv).dimensions is length


def test_decagram():
    dag = Unit('dag')
    g = Unit('g')
    assert dag.get_conversion_factor(g) == (10.0, None)


def test_pickle():
    cm = Unit('cm')
    assert cm == pickle.loads(pickle.dumps(cm))


def test_preserve_offset():
    from unyt import degF, dimensionless

    new_unit = degF*dimensionless

    assert new_unit is not degF
    assert new_unit == degF
    assert new_unit.base_offset == degF.base_offset

    new_unit = degF/dimensionless

    assert new_unit is not degF
    assert new_unit == degF
    assert new_unit.base_offset == degF.base_offset

    with pytest.raises(InvalidUnitOperation):
        dimensionless/degF


def test_code_unit():
    from unyt import UnitRegistry

    ureg = UnitRegistry()
    ureg.add('code_length', 10., length)
    ureg.add('code_magnetic_field', 2.0, magnetic_field_mks)
    u = Unit('code_length', registry=ureg)
    assert u.is_code_unit is True
    assert u.get_base_equivalent() == Unit('m')
    u = Unit('cm')
    assert u.is_code_unit is False

    u = Unit('code_magnetic_field', registry=ureg)
    assert u.get_base_equivalent('mks') == Unit('T')
    assert u.get_base_equivalent('cgs') == Unit('gauss')

    UnitSystem(ureg.unit_system_id, 'code_length', 'kg', 's', registry=ureg)

    u = Unit('cm', registry=ureg)
    ue = u.get_base_equivalent('code')

    assert str(ue) == 'code_length'
    assert ue.base_value == 10
    assert ue.dimensions is length

    class FakeDataset(object):
        unit_registry = ureg

    ds = FakeDataset()

    UnitSystem(ds, 'code_length', 'kg', 's', registry=ureg)

    u = Unit('cm', registry=ureg)
    ue = u.get_base_equivalent(ds)

    assert str(ue) == 'code_length'
    assert ue.base_value == 10
    assert ue.dimensions is length

    with pytest.raises(UnitParseError):
        Unit('code_length')


def test_bad_equivalence():
    from unyt import cm

    with pytest.raises(KeyError):
        cm.has_equivalent('dne')


def test_em_unit_base_equivalent():
    from unyt import A, cm

    with pytest.raises(UnitsNotReducible):
        (A/cm).get_base_equivalent('cgs')


def test_define_unit_error():
    from unyt import define_unit

    with pytest.raises(RuntimeError):
        define_unit('foobar', 'baz')
    with pytest.raises(RuntimeError):
        define_unit('foobar', 12)


def test_symbol_lut_length():
    for v in default_unit_symbol_lut.values():
        assert len(v) == 5



"""
Test ndarray subclass that handles symbolic units.




"""

# ----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

import copy
from six.moves import cPickle as pickle
import itertools
import math
import numpy as np
import operator
import os
import pytest
import shutil
import tempfile

from numpy.testing import (
    assert_array_equal,
    assert_equal,
    assert_array_almost_equal_nulp,
    assert_array_almost_equal,
    assert_almost_equal
)
from numpy import array
from unyt.array import (
    unyt_array,
    unyt_quantity,
    unary_operators,
    binary_operators,
    uconcatenate,
    ucross,
    udot,
    uintersect1d,
    unorm,
    uunion1d,
    loadtxt,
    savetxt
)
from unyt.exceptions import (
    InvalidUnitEquivalence,
    InvalidUnitOperation,
    IterableUnitCoercionError,
    UnitConversionError,
    UnitOperationError,
    UnitParseError,
    UnitsNotReducible,
)
from unyt._testing import assert_allclose_units
from unyt.unit_symbols import (
    cm,
    m,
    g,
    degree
)
from unyt.unit_registry import UnitRegistry
from unyt._on_demand_imports import (
    _astropy,
    _h5py,
    _pint,
    NotAModule,
)
from unyt._physical_ratios import (
    metallicity_sun,
    speed_of_light_cm_per_s
)
from unyt import dimensions


def operate_and_compare(a, b, op, answer):
    # Test generator for unyt_arrays tests
    assert_array_equal(op(a, b), answer)


def assert_isinstance(a, type):
    assert isinstance(a, type)


def test_addition():
    """
    Test addition of two unyt_arrays

    """

    # Same units
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'cm')
    a3 = [4*cm, 5*cm, 6*cm]
    answer = unyt_array([5, 7, 9], 'cm')

    operate_and_compare(a1, a2, operator.add, answer)
    operate_and_compare(a2, a1, operator.add, answer)
    operate_and_compare(a1, a3, operator.add, answer)
    operate_and_compare(a3, a1, operator.add, answer)
    operate_and_compare(a2, a1, np.add, answer)
    operate_and_compare(a1, a2, np.add, answer)
    operate_and_compare(a1, a3, np.add, answer)
    operate_and_compare(a3, a1, np.add, answer)

    # different units
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'm')
    a3 = [4*m, 5*m, 6*m]
    answer1 = unyt_array([401, 502, 603], 'cm')
    answer2 = unyt_array([4.01, 5.02, 6.03], 'm')

    operate_and_compare(a1, a2, operator.add, answer1)
    operate_and_compare(a2, a1, operator.add, answer2)
    operate_and_compare(a1, a3, operator.add, answer1)
    operate_and_compare(a3, a1, operator.add, answer2)
    operate_and_compare(a1, a2, np.add, answer1)
    operate_and_compare(a2, a1, np.add, answer2)
    operate_and_compare(a1, a3, np.add, answer1)
    operate_and_compare(a3, a1, np.add, answer2)

    # Test dimensionless quantities
    a1 = unyt_array([1, 2, 3])
    a2 = array([4, 5, 6])
    a3 = [4, 5, 6]
    answer = unyt_array([5, 7, 9])

    operate_and_compare(a1, a2, operator.add, answer)
    operate_and_compare(a2, a1, operator.add, answer)
    operate_and_compare(a1, a3, operator.add, answer)
    operate_and_compare(a3, a1, operator.add, answer)
    operate_and_compare(a1, a2, np.add, answer)
    operate_and_compare(a2, a1, np.add, answer)
    operate_and_compare(a1, a3, np.add, answer)
    operate_and_compare(a3, a1, np.add, answer)

    # Catch the different dimensions error
    a1 = unyt_array([1, 2, 3], 'm')
    a2 = unyt_array([4, 5, 6], 'kg')
    a3 = [7, 8, 9]
    a4 = unyt_array([10, 11, 12], '')

    with pytest.raises(UnitOperationError):
        operator.add(a1, a2)
    with pytest.raises(UnitOperationError):
        operator.iadd(a1, a2)
    with pytest.raises(UnitOperationError):
        operator.add(a1, a3)
    with pytest.raises(UnitOperationError):
        operator.iadd(a1, a3)
    with pytest.raises(UnitOperationError):
        operator.add(a3, a1)
    with pytest.raises(UnitOperationError):
        operator.iadd(a3, a1)
    with pytest.raises(UnitOperationError):
        operator.add(a1, a4)
    with pytest.raises(UnitOperationError):
        operator.iadd(a1, a4)
    with pytest.raises(UnitOperationError):
        operator.add(a4, a1)
    with pytest.raises(UnitOperationError):
        operator.iadd(a4, a1)

    # adding with zero is allowed irrespective of the units
    zeros = np.zeros(3)
    zeros_yta_dimless = unyt_array(zeros, 'dimensionless')
    zeros_yta_length = unyt_array(zeros, 'm')
    zeros_yta_mass = unyt_array(zeros, 'kg')
    operands = [0, zeros, zeros_yta_length]

    for op in [operator.add, np.add]:
        for operand in operands:
            operate_and_compare(a1, operand, op, a1)
            operate_and_compare(operand, a1, op, a1)
            operate_and_compare(4*m, operand, op, 4*m)
            operate_and_compare(operand, 4*m, op, 4*m)

    operands = [unyt_quantity(0), unyt_quantity(0, 'kg'),
                zeros_yta_dimless, zeros_yta_mass]

    for op in [operator.add, np.add]:
        for operand in operands:
            with pytest.raises(UnitOperationError):
                operate_and_compare(a1, operand, op, a1)
            with pytest.raises(UnitOperationError):
                operate_and_compare(operand, a1, op, a1)
            with pytest.raises(UnitOperationError):
                operate_and_compare(4*m, operand, op, 4*m)
            with pytest.raises(UnitOperationError):
                operate_and_compare(operand, 4*m, op, 4*m)


def test_subtraction():
    """
    Test subtraction of two unyt_arrays

    """

    # Same units
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'cm')
    a3 = [4*cm, 5*cm, 6*cm]
    answer1 = unyt_array([-3, -3, -3], 'cm')
    answer2 = unyt_array([3, 3, 3], 'cm')

    operate_and_compare(a1, a2, operator.sub, answer1)
    operate_and_compare(a2, a1, operator.sub, answer2)
    operate_and_compare(a1, a3, operator.sub, answer1)
    operate_and_compare(a3, a1, operator.sub, answer2)
    operate_and_compare(a1, a2, np.subtract, answer1)
    operate_and_compare(a2, a1, np.subtract, answer2)
    operate_and_compare(a1, a3, np.subtract, answer1)
    operate_and_compare(a3, a1, np.subtract, answer2)

    # different units
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'm')
    a3 = [4*m, 5*m, 6*m]
    answer1 = unyt_array([-399, -498, -597], 'cm')
    answer2 = unyt_array([3.99, 4.98, 5.97], 'm')
    answer3 = unyt_array([399, 498, 597], 'cm')

    operate_and_compare(a1, a2, operator.sub, answer1)
    operate_and_compare(a2, a1, operator.sub, answer2)
    operate_and_compare(a1, a3, operator.sub, answer1)
    operate_and_compare(a3, a1, operator.sub, answer3)
    operate_and_compare(a1, a2, np.subtract, answer1)
    operate_and_compare(a2, a1, np.subtract, answer2)
    operate_and_compare(a1, a3, np.subtract, answer1)
    operate_and_compare(a3, a1, np.subtract, answer3)

    # Test dimensionless quantities
    a1 = unyt_array([1, 2, 3])
    a2 = array([4, 5, 6])
    a3 = [4, 5, 6]
    answer1 = unyt_array([-3, -3, -3])
    answer2 = unyt_array([3, 3, 3])

    operate_and_compare(a1, a2, operator.sub, answer1)
    operate_and_compare(a2, a1, operator.sub, answer2)
    operate_and_compare(a1, a3, operator.sub, answer1)
    operate_and_compare(a3, a1, operator.sub, answer2)
    operate_and_compare(a1, a2, np.subtract, answer1)
    operate_and_compare(a2, a1, np.subtract, answer2)
    operate_and_compare(a1, a3, np.subtract, answer1)
    operate_and_compare(a3, a1, np.subtract, answer2)

    # Catch the different dimensions error
    a1 = unyt_array([1, 2, 3], 'm')
    a2 = unyt_array([4, 5, 6], 'kg')
    a3 = [7, 8, 9]
    a4 = unyt_array([10, 11, 12], '')

    with pytest.raises(UnitOperationError):
        operator.sub(a1, a2)
    with pytest.raises(UnitOperationError):
        operator.isub(a1, a2)
    with pytest.raises(UnitOperationError):
        operator.sub(a1, a3)
    with pytest.raises(UnitOperationError):
        operator.isub(a1, a3)
    with pytest.raises(UnitOperationError):
        operator.sub(a3, a1)
    with pytest.raises(UnitOperationError):
        operator.isub(a3, a1)
    with pytest.raises(UnitOperationError):
        operator.sub(a1, a4)
    with pytest.raises(UnitOperationError):
        operator.isub(a1, a4)
    with pytest.raises(UnitOperationError):
        operator.sub(a4, a1)
    with pytest.raises(UnitOperationError):
        operator.isub(a4, a1)

    # subtracting with zero is allowed irrespective of the units
    zeros = np.zeros(3)
    zeros_yta_dimless = unyt_array(zeros, 'dimensionless')
    zeros_yta_length = unyt_array(zeros, 'm')
    zeros_yta_mass = unyt_array(zeros, 'kg')
    operands = [0, zeros, zeros_yta_length]

    for op in [operator.sub, np.subtract]:
        for operand in operands:
            operate_and_compare(a1, operand, op, a1)
            operate_and_compare(operand, a1, op, -a1)
            operate_and_compare(4*m, operand, op, 4*m)
            operate_and_compare(operand, 4*m, op, -4*m)

    operands = [unyt_quantity(0), unyt_quantity(0, 'kg'),
                zeros_yta_dimless, zeros_yta_mass]

    for op in [operator.sub, np.subtract]:
        for operand in operands:
            with pytest.raises(UnitOperationError):
                operate_and_compare(a1, operand, op, a1)
            with pytest.raises(UnitOperationError):
                operate_and_compare(operand, a1, op, -a1)
            with pytest.raises(UnitOperationError):
                operate_and_compare(4*m, operand, op, 4*m)
            with pytest.raises(UnitOperationError):
                operate_and_compare(operand, 4*m, op, -4*m)


def test_multiplication():
    """
    Test multiplication of two unyt_arrays

    """

    # Same units
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'cm')
    a3 = [4*cm, 5*cm, 6*cm]
    answer = unyt_array([4, 10, 18], 'cm**2')

    operate_and_compare(a1, a2, operator.mul, answer)
    operate_and_compare(a2, a1, operator.mul, answer)
    operate_and_compare(a1, a3, operator.mul, answer)
    operate_and_compare(a3, a1, operator.mul, answer)
    operate_and_compare(a1, a2, np.multiply, answer)
    operate_and_compare(a2, a1, np.multiply, answer)
    operate_and_compare(a1, a3, np.multiply, answer)
    operate_and_compare(a3, a1, np.multiply, answer)

    # different units, same dimension
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'm')
    a3 = [4*m, 5*m, 6*m]
    answer1 = unyt_array([400, 1000, 1800], 'cm**2')
    answer2 = unyt_array([.04, .10, .18], 'm**2')
    answer3 = unyt_array([4, 10, 18], 'cm*m')

    operate_and_compare(a1, a2, operator.mul, answer1)
    operate_and_compare(a2, a1, operator.mul, answer2)
    operate_and_compare(a1, a3, operator.mul, answer1)
    operate_and_compare(a3, a1, operator.mul, answer2)
    operate_and_compare(a1, a2, np.multiply, answer3)
    operate_and_compare(a2, a1, np.multiply, answer3)
    operate_and_compare(a1, a3, np.multiply, answer3)
    operate_and_compare(a3, a1, np.multiply, answer3)

    # different dimensions
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([4, 5, 6], 'g')
    a3 = [4*g, 5*g, 6*g]
    answer = unyt_array([4, 10, 18], 'cm*g')

    operate_and_compare(a1, a2, operator.mul, answer)
    operate_and_compare(a2, a1, operator.mul, answer)
    operate_and_compare(a1, a3, operator.mul, answer)
    operate_and_compare(a3, a1, operator.mul, answer)
    operate_and_compare(a1, a2, np.multiply, answer)
    operate_and_compare(a2, a1, np.multiply, answer)
    operate_and_compare(a1, a3, np.multiply, answer)
    operate_and_compare(a3, a1, np.multiply, answer)

    # One dimensionless, one unitful
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = array([4, 5, 6])
    a3 = [4, 5, 6]
    answer = unyt_array([4, 10, 18], 'cm')

    operate_and_compare(a1, a2, operator.mul, answer)
    operate_and_compare(a2, a1, operator.mul, answer)
    operate_and_compare(a1, a3, operator.mul, answer)
    operate_and_compare(a3, a1, operator.mul, answer)
    operate_and_compare(a1, a2, np.multiply, answer)
    operate_and_compare(a2, a1, np.multiply, answer)
    operate_and_compare(a1, a3, np.multiply, answer)
    operate_and_compare(a3, a1, np.multiply, answer)

    # Both dimensionless quantities
    a1 = unyt_array([1, 2, 3])
    a2 = array([4, 5, 6])
    a3 = [4, 5, 6]
    answer = unyt_array([4, 10, 18])

    operate_and_compare(a1, a2, operator.mul, answer)
    operate_and_compare(a2, a1, operator.mul, answer)
    operate_and_compare(a1, a3, operator.mul, answer)
    operate_and_compare(a3, a1, operator.mul, answer)
    operate_and_compare(a1, a2, np.multiply, answer)
    operate_and_compare(a2, a1, np.multiply, answer)
    operate_and_compare(a1, a3, np.multiply, answer)
    operate_and_compare(a3, a1, np.multiply, answer)


def test_division():
    """
    Test multiplication of two unyt_arrays

    """

    # Same units
    a1 = unyt_array([1., 2., 3.], 'cm')
    a2 = unyt_array([4., 5., 6.], 'cm')
    a3 = [4*cm, 5*cm, 6*cm]
    answer1 = unyt_array([0.25, 0.4, 0.5])
    answer2 = unyt_array([4, 2.5, 2])
    if "div" in dir(operator):
        op = operator.div
    else:
        op = operator.truediv

    operate_and_compare(a1, a2, op, answer1)
    operate_and_compare(a2, a1, op, answer2)
    operate_and_compare(a1, a3, op, answer1)
    operate_and_compare(a3, a1, op, answer2)
    operate_and_compare(a1, a2, np.divide, answer1)
    operate_and_compare(a2, a1, np.divide, answer2)
    operate_and_compare(a1, a3, np.divide, answer1)
    operate_and_compare(a3, a1, np.divide, answer2)

    # different units, same dimension
    a1 = unyt_array([1., 2., 3.], 'cm')
    a2 = unyt_array([4., 5., 6.], 'm')
    a3 = [4*m, 5*m, 6*m]
    answer1 = unyt_array([.0025, .004, .005])
    answer2 = unyt_array([400, 250, 200])

    operate_and_compare(a1, a2, op, answer1)
    operate_and_compare(a2, a1, op, answer2)
    operate_and_compare(a1, a3, op, answer1)
    operate_and_compare(a3, a1, op, answer2)
    operate_and_compare(a1, a2, np.divide, answer1)
    operate_and_compare(a2, a1, np.divide, answer2)
    operate_and_compare(a1, a3, np.divide, answer1)
    operate_and_compare(a3, a1, np.divide, answer2)

    # different dimensions
    a1 = unyt_array([1., 2., 3.], 'cm')
    a2 = unyt_array([4., 5., 6.], 'g')
    a3 = [4*g, 5*g, 6*g]
    answer1 = unyt_array([0.25, 0.4, 0.5], 'cm/g')
    answer2 = unyt_array([4, 2.5, 2], 'g/cm')

    operate_and_compare(a1, a2, op, answer1)
    operate_and_compare(a2, a1, op, answer2)
    operate_and_compare(a1, a3, op, answer1)
    operate_and_compare(a3, a1, op, answer2)
    operate_and_compare(a1, a2, np.divide, answer1)
    operate_and_compare(a2, a1, np.divide, answer2)
    operate_and_compare(a1, a3, np.divide, answer1)
    operate_and_compare(a3, a1, np.divide, answer2)

    # One dimensionless, one unitful
    a1 = unyt_array([1., 2., 3.], 'cm')
    a2 = array([4., 5., 6.])
    a3 = [4, 5, 6]
    answer1 = unyt_array([0.25, 0.4, 0.5], 'cm')
    answer2 = unyt_array([4, 2.5, 2], '1/cm')

    operate_and_compare(a1, a2, op, answer1)
    operate_and_compare(a2, a1, op, answer2)
    operate_and_compare(a1, a3, op, answer1)
    operate_and_compare(a3, a1, op, answer2)
    operate_and_compare(a1, a2, np.divide, answer1)
    operate_and_compare(a2, a1, np.divide, answer2)
    operate_and_compare(a1, a3, np.divide, answer1)
    operate_and_compare(a3, a1, np.divide, answer2)

    # Both dimensionless quantities
    a1 = unyt_array([1., 2., 3.])
    a2 = array([4., 5., 6.])
    a3 = [4, 5, 6]
    answer1 = unyt_array([0.25, 0.4, 0.5])
    answer2 = unyt_array([4, 2.5, 2])

    operate_and_compare(a1, a2, op, answer1)
    operate_and_compare(a2, a1, op, answer2)
    operate_and_compare(a1, a3, op, answer1)
    operate_and_compare(a3, a1, op, answer2)
    operate_and_compare(a1, a3, np.divide, answer1)
    operate_and_compare(a3, a1, np.divide, answer2)
    operate_and_compare(a1, a3, np.divide, answer1)
    operate_and_compare(a3, a1, np.divide, answer2)


def test_power():
    """
    Test power operator ensure units are correct.

    """

    from unyt import cm

    cm_arr = np.array([1.0, 1.0]) * cm

    assert_equal((1*cm)**3, unyt_quantity(1, 'cm**3'))
    assert_equal(np.power((1*cm), 3), unyt_quantity(1, 'cm**3'))
    assert_equal((1*cm)**unyt_quantity(3), unyt_quantity(1, 'cm**3'))
    with pytest.raises(UnitOperationError):
        np.power((1*cm), unyt_quantity(3, 'g'))
    with pytest.raises(InvalidUnitOperation):
        np.power(cm, cm)

    assert_equal(cm_arr**3, unyt_array([1, 1], 'cm**3'))
    assert_equal(np.power(cm_arr, 3), unyt_array([1, 1], 'cm**3'))
    assert_equal(cm_arr**unyt_quantity(3), unyt_array([1, 1], 'cm**3'))
    with pytest.raises(UnitOperationError):
        np.power(cm_arr, unyt_quantity(3, 'g'))


def test_comparisons():
    """
    Test numpy ufunc comparison operators for unit consistency.

    """
    from unyt.array import unyt_array

    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([2, 1, 3], 'cm')
    a3 = unyt_array([.02, .01, .03], 'm')
    dimless = np.array([2, 1, 3])

    ops = (
        np.less,
        np.less_equal,
        np.greater,
        np.greater_equal,
        np.equal,
        np.not_equal
    )

    answers = (
        [True, False, False],
        [True, False, True],
        [False, True, False],
        [False, True, True],
        [False, False, True],
        [True, True, False],
    )

    for op, answer in zip(ops, answers):
        operate_and_compare(a1, a2, op, answer)
    for op, answer in zip(ops, answers):
        operate_and_compare(a1, dimless, op, answer)

    for op, answer in zip(ops, answers):
        operate_and_compare(a1, a3, op, answer)

    for op, answer in zip(ops, answers):
        operate_and_compare(a1, a3.in_units('cm'), op, answer)

    # Check that comparisons with dimensionless quantities work in both
    # directions.
    operate_and_compare(a3, dimless, np.less, [True, True, True])
    operate_and_compare(dimless, a3, np.less, [False, False, False])
    assert_equal(a1 < 2, [True, False, False])
    assert_equal(a1 < 2, np.less(a1, 2))
    assert_equal(2 < a1, [False, False, True])
    assert_equal(2 < a1, np.less(2, a1))


def test_unit_conversions():
    """
    Test operations that convert to different units or cast to ndarray

    """
    from unyt.array import unyt_quantity
    from unyt.unit_object import Unit

    km = unyt_quantity(1, 'km')
    km_in_cm = km.in_units('cm')
    cm_unit = Unit('cm')
    kpc_unit = Unit('kpc')

    assert_equal(km_in_cm, km)
    assert_equal(km_in_cm.in_cgs(), 1e5)
    assert_equal(km_in_cm.in_mks(), 1e3)
    assert_equal(km_in_cm.units, cm_unit)

    km_view = km.ndarray_view()
    km.convert_to_units('cm')
    assert km_view.base is km.base

    assert_equal(km, unyt_quantity(1, 'km'))
    assert_equal(km.in_cgs(), 1e5)
    assert_equal(km.in_mks(), 1e3)
    assert_equal(km.units, cm_unit)

    km.convert_to_units('kpc')
    assert km_view.base is km.base

    assert_array_almost_equal_nulp(km, unyt_quantity(1, 'km'))
    assert_array_almost_equal_nulp(km.in_cgs(), unyt_quantity(1e5, 'cm'))
    assert_array_almost_equal_nulp(km.in_mks(), unyt_quantity(1e3, 'm'))
    assert_equal(km.units, kpc_unit)

    assert_isinstance(km.to_ndarray(), np.ndarray)
    assert_isinstance(km.ndarray_view(), np.ndarray)

    dyne = unyt_quantity(1.0, 'dyne')

    assert_equal(dyne.in_cgs(), dyne)
    assert_equal(dyne.in_cgs(), 1.0)
    assert_equal(dyne.in_mks(), dyne)
    assert_equal(dyne.in_mks(), 1e-5)
    assert_equal(str(dyne.in_mks().units), 'N')
    assert_equal(str(dyne.in_cgs().units), 'dyne')

    em3 = unyt_quantity(1.0, 'erg/m**3')

    assert_equal(em3.in_cgs(), em3)
    assert_equal(em3.in_cgs(), 1e-6)
    assert_equal(em3.in_mks(), em3)
    assert_equal(em3.in_mks(), 1e-7)
    assert_equal(str(em3.in_mks().units), 'Pa')
    assert_equal(str(em3.in_cgs().units), 'dyne/cm**2')

    em3_converted = unyt_quantity(1545436840.386756, 'Msun/(Myr**2*kpc)')
    assert_equal(em3.in_base(unit_system="galactic"), em3)
    assert_array_almost_equal(
        em3.in_base(unit_system="galactic"), em3_converted)
    assert_equal(
        str(em3.in_base(unit_system="galactic").units), 'Msun/(Myr**2*kpc)')

    dimless = unyt_quantity(1.0, "")
    assert_equal(dimless.in_cgs(), dimless)
    assert_equal(dimless.in_cgs(), 1.0)
    assert_equal(dimless.in_mks(), dimless)
    assert_equal(dimless.in_mks(), 1.0)
    assert_equal(str(dimless.in_cgs().units), "dimensionless")

    kg = unyt_quantity(1.0, 'kg')
    assert kg.to(g).v == 1000
    assert kg.in_units(g).v == 1000
    kg.convert_to_units(g)
    assert kg.v == 1000

    ten_grams = 10*g
    assert kg.to(ten_grams).v == 100
    assert kg.in_units(ten_grams).v == 100
    kg.convert_to_units(ten_grams)
    assert kg.v == 100

    with pytest.raises(UnitParseError):
        kg.to([1, 2]*g)

    with pytest.raises(UnitParseError):
        kg.in_units([1, 2]*g)

    with pytest.raises(UnitParseError):
        kg.convert_to_units([1, 2]*g)


def test_temperature_conversions():
    """
    Test conversions between various supported temperatue scales.

    Also ensure we only allow compound units with temperature
    scales that have a proper zero point.

    """
    from unyt.unit_object import InvalidUnitOperation

    km = unyt_quantity(1, 'km')
    balmy = unyt_quantity(300, 'K')
    balmy_F = unyt_quantity(80.33, 'degF')
    balmy_C = unyt_quantity(26.85, 'degC')
    balmy_R = unyt_quantity(540, 'R')

    assert_array_almost_equal(balmy.in_units('degF').d, balmy_F.d)
    assert balmy.in_units('degF').units, balmy_F.units
    assert_array_almost_equal(balmy.in_units('degC').d, balmy_C.d)
    assert balmy.in_units('degC').units, balmy_C.units
    assert_array_almost_equal(balmy.in_units('R').d, balmy_R.d)
    assert balmy.in_units('R').units == balmy_R.units

    balmy_view = balmy.ndarray_view()

    balmy.convert_to_units('degF')
    assert balmy_view.base is balmy.base
    assert_array_almost_equal(np.array(balmy), np.array(balmy_F))

    balmy.convert_to_units('degC')
    assert balmy_view.base is balmy.base
    assert_array_almost_equal(np.array(balmy), np.array(balmy_C))

    balmy.convert_to_units('R')
    assert balmy_view.base is balmy.base
    assert_array_almost_equal(np.array(balmy), np.array(balmy_R))

    balmy.convert_to_units('degF')
    assert balmy_view.base is balmy.base
    assert_array_almost_equal(np.array(balmy), np.array(balmy_F))

    with pytest.raises(InvalidUnitOperation):
        np.multiply(balmy, km)
    with pytest.raises(InvalidUnitOperation):
        np.multiply(balmy, balmy)
    with pytest.raises(InvalidUnitOperation):
        np.multiply(balmy_F, balmy_F)
    with pytest.raises(InvalidUnitOperation):
        np.multiply(balmy_F, balmy_C)
    with pytest.raises(InvalidUnitOperation):
        np.divide(balmy, balmy)
    with pytest.raises(InvalidUnitOperation):
        np.divide(balmy_F, balmy_F)
    with pytest.raises(InvalidUnitOperation):
        np.divide(balmy_F, balmy_C)
    with pytest.raises(InvalidUnitOperation):
        balmy*km
    with pytest.raises(InvalidUnitOperation):
        balmy*balmy
    with pytest.raises(InvalidUnitOperation):
        balmy_F*balmy_F
    with pytest.raises(InvalidUnitOperation):
        balmy_F*balmy_C
    with pytest.raises(InvalidUnitOperation):
        2*balmy_F
    with pytest.raises(InvalidUnitOperation):
        balmy/balmy
    with pytest.raises(InvalidUnitOperation):
        balmy_F/balmy_F
    with pytest.raises(InvalidUnitOperation):
        balmy_F/balmy_C
    assert np.add(balmy_F, balmy_F) == unyt_quantity(80.33*2, 'degF')
    with pytest.raises(InvalidUnitOperation):
        np.add(balmy_F, balmy_C)
    with pytest.raises(InvalidUnitOperation):
        balmy_F + balmy_C
    assert_equal(np.subtract(balmy_C, balmy_C), unyt_quantity(0, 'degC'))
    with pytest.raises(InvalidUnitOperation):
        np.subtract(balmy_F, balmy_C)
    with pytest.raises(InvalidUnitOperation):
        balmy_F - balmy_C

    # Does CGS conversion from F to K work?
    assert_array_almost_equal(balmy.in_cgs(), unyt_quantity(300, 'K'))


def test_unyt_array_unyt_quantity_ops():
    """
    Test operations that combine unyt_array and unyt_quantity
    """
    a = unyt_array(range(10, 1), 'cm')
    b = unyt_quantity(5, 'g')

    assert_isinstance(a*b, unyt_array)
    assert_isinstance(b*a, unyt_array)

    assert_isinstance(a/b, unyt_array)
    assert_isinstance(b/a, unyt_array)

    assert_isinstance(a*a, unyt_array)
    assert_isinstance(a/a, unyt_array)

    assert_isinstance(b*b, unyt_quantity)
    assert_isinstance(b/b, unyt_quantity)


def test_selecting():
    """
    Test slicing of two unyt_arrays

    """
    a = unyt_array(range(10), 'cm')
    a_slice = a[:3]
    a_fancy_index = a[[1, 1, 3, 5]]
    a_array_fancy_index = a[array([[1, 1], [3, 5]])]
    a_boolean_index = a[a > 5]
    a_selection = a[0]

    assert_array_equal(a_slice, unyt_array([0, 1, 2], 'cm'))
    assert_equal(a_slice.units, a.units)
    assert_array_equal(a_fancy_index, unyt_array([1, 1, 3, 5], 'cm'))
    assert_equal(a_fancy_index.units, a.units)
    assert_array_equal(a_array_fancy_index,
                       unyt_array([[1, 1, ], [3, 5]], 'cm'))
    assert_equal(a_array_fancy_index.units, a.units)
    assert_array_equal(a_boolean_index, unyt_array([6, 7, 8, 9], 'cm'))
    assert_equal(a_boolean_index.units, a.units)
    assert_isinstance(a_selection, unyt_quantity)
    assert_equal(a_selection.units, a.units)

    # .base points to the original array for a numpy view.  If it is not a
    # view, .base is None.
    assert a_slice.base is a


def test_iteration():
    """
    Test that iterating over a unyt_array returns a sequence of unyt_quantity
    instances
    """
    a = np.arange(3)
    b = unyt_array(np.arange(3), 'cm')
    for ia, ib, in zip(a, b):
        assert_equal(ia, ib.value)
        assert_equal(ib.units, b.units)


def test_ytarray_pickle():
    test_data = [unyt_quantity(12.0, 'cm'),
                 unyt_array([1, 2, 3], 'km')]

    for data in test_data:
        tempf = tempfile.NamedTemporaryFile(delete=False)
        pickle.dump(data, tempf)
        tempf.close()

        with open(tempf.name, "rb") as fname:
            loaded_data = pickle.load(fname)
        os.unlink(tempf.name)

        assert_array_equal(data, loaded_data)
        assert_equal(data.units, loaded_data.units)
        assert_array_equal(array(data.in_cgs()), array(loaded_data.in_cgs()))
        assert_equal(float(data.units.base_value),
                     float(loaded_data.units.base_value))


def test_copy():
    quan = unyt_quantity(1, 'g')
    arr = unyt_array([1, 2, 3], 'cm')

    assert_equal(copy.copy(quan), quan)
    assert_array_equal(copy.copy(arr), arr)

    assert_equal(copy.deepcopy(quan), quan)
    assert_array_equal(copy.deepcopy(arr), arr)

    memo = {}
    assert_equal(copy.deepcopy(quan, memo), quan)
    assert_array_equal(copy.deepcopy(arr), arr)

    assert_equal(quan.copy(), quan)
    assert_array_equal(arr.copy(), arr)

    assert_equal(np.copy(quan), quan)
    assert_array_equal(np.copy(arr), arr)


# needed so the tests function on older numpy versions that have
# different sets of ufuncs
def yield_np_ufuncs(ufunc_list):
    for u in ufunc_list:
        ufunc = getattr(np, u, None)
        if ufunc is not None:
            yield ufunc


def unary_ufunc_comparison(ufunc, a):
    out = a.copy()
    a_array = a.to_ndarray()
    if ufunc in (np.isreal, np.iscomplex, ):
        # According to the numpy docs, these two explicitly do not do
        # in-place copies.
        ret = ufunc(a)
        assert not hasattr(ret, 'units')
        assert_array_equal(ret, ufunc(a))
    elif ufunc in yield_np_ufuncs([
            'exp', 'exp2', 'log', 'log2', 'log10', 'expm1', 'log1p', 'sin',
            'cos', 'tan', 'arcsin', 'arccos', 'arctan', 'sinh', 'cosh', 'tanh',
            'arccosh', 'arcsinh', 'arctanh', 'deg2rad', 'rad2deg', 'isfinite',
            'isinf', 'isnan', 'signbit', 'sign', 'rint', 'logical_not']):
        # These operations should return identical results compared to numpy.
        with np.errstate(invalid='ignore'):
            ret = ufunc(a, out=out)

            assert_array_equal(ret, out)
            assert_array_equal(ret, ufunc(a_array))
            # In-place copies do not drop units.
            assert hasattr(out, 'units')
            assert not hasattr(ret, 'units')
    elif ufunc in yield_np_ufuncs(
            ['absolute', 'fabs', 'conjugate', 'floor', 'ceil', 'trunc',
             'negative', 'spacing', 'positive']):

        ret = ufunc(a, out=out)

        assert_array_equal(ret, out)
        assert_array_equal(ret.to_ndarray(), ufunc(a_array))
        assert ret.units == out.units
    elif ufunc in yield_np_ufuncs(
            ['ones_like', 'square', 'sqrt', 'reciprocal']):
        if ufunc is np.ones_like:
            ret = ufunc(a)
        else:
            with np.errstate(invalid='ignore'):
                ret = ufunc(a, out=out)
            assert_array_equal(ret, out)

        with np.errstate(invalid='ignore'):
            assert_array_equal(ret.to_ndarray(), ufunc(a_array))
        if ufunc is np.square:
            assert out.units == a.units**2
            assert ret.units == a.units**2
        elif ufunc is np.sqrt:
            assert out.units == a.units**0.5
            assert ret.units == a.units**0.5
        elif ufunc is np.reciprocal:
            assert out.units == a.units**-1
            assert ret.units == a.units**-1
    elif ufunc is np.modf:
        ret1, ret2 = ufunc(a)
        npret1, npret2 = ufunc(a_array)

        assert_array_equal(ret1.to_ndarray(), npret1)
        assert_array_equal(ret2.to_ndarray(), npret2)
    elif ufunc is np.frexp:
        ret1, ret2 = ufunc(a)
        npret1, npret2 = ufunc(a_array)

        assert_array_equal(ret1, npret1)
        assert_array_equal(ret2, npret2)
    elif ufunc is np.invert:
        with pytest.raises(TypeError):
            ufunc(a.astype('int64'))
    elif hasattr(np, 'isnat') and ufunc is np.isnat:
        # numpy 1.13 raises ValueError, numpy 1.14 and newer raise TypeError
        with pytest.raises((TypeError, ValueError)):
            ufunc(a)
    # no untested ufuncs
    assert ufunc in yield_np_ufuncs([
        'isreal', 'iscomplex', 'exp', 'exp2', 'log', 'log2', 'log10', 'expm1',
        'log1p', 'sin', 'cos', 'tan', 'arcsin', 'arccos', 'arctan', 'sinh',
        'cosh', 'tanh', 'arccosh', 'arcsinh', 'arctanh', 'deg2rad', 'rad2deg',
        'isfinite', 'isinf', 'isnan', 'signbit', 'sign', 'rint', 'logical_not',
        'absolute', 'fabs', 'conjugate', 'floor', 'ceil', 'trunc', 'negative',
        'spacing', 'positive', 'ones_like', 'square', 'sqrt', 'reciprocal',
        'invert', 'isnat', 'modf', 'frexp',
    ])


def binary_ufunc_comparison(ufunc, a, b):
    out = b.copy()
    if ufunc in yield_np_ufuncs([
            'add', 'subtract', 'remainder', 'fmod', 'mod', 'arctan2', 'hypot',
            'greater', 'greater_equal', 'less', 'less_equal', 'equal',
            'not_equal', 'logical_and', 'logical_or', 'logical_xor', 'maximum',
            'minimum', 'fmax', 'fmin', 'nextafter', 'heaviside']):
        if a.units != b.units and a.units.dimensions != b.units.dimensions:
            with pytest.raises(UnitOperationError):
                ufunc(a, b)
            return
    if ufunc in yield_np_ufuncs(
            ['bitwise_and', 'bitwise_or', 'bitwise_xor', 'left_shift',
             'right_shift', 'ldexp']):
        with pytest.raises(TypeError):
            ufunc(a, b)
        return

    ret = ufunc(a, b, out=out)
    ret = ufunc(a, b)

    if ufunc is np.multiply:
        assert ret.units == a.units*b.units
    elif ufunc in (np.divide, np.true_divide, np.arctan2):
        assert ret.units.dimensions == (a.units/b.units).dimensions
    elif ufunc in (np.greater, np.greater_equal, np.less, np.less_equal,
                   np.not_equal, np.equal, np.logical_and, np.logical_or,
                   np.logical_xor):
        assert not isinstance(ret, unyt_array) and isinstance(ret, np.ndarray)
    if isinstance(ret, tuple):
        assert_array_equal(ret[0], out)
    else:
        assert_array_equal(ret, out)
    if ((ufunc in (np.divide, np.true_divide, np.arctan2) and
         (a.units.dimensions == b.units.dimensions))):
        assert_array_almost_equal(
            np.array(ret), ufunc(np.array(a.in_cgs()), np.array(b.in_cgs())))


def test_ufuncs():
    for ufunc in unary_operators:
        unary_ufunc_comparison(ufunc, unyt_array([.3, .4, .5], 'cm'))
        unary_ufunc_comparison(ufunc, unyt_array([12, 23, 47], 'g'))
        unary_ufunc_comparison(ufunc, unyt_array([2, 4, -6], 'erg/m**3'))

    for ufunc in binary_operators:
        # arr**arr is undefined for arrays with units because
        # each element of the result would have different units.
        if ufunc is np.power:
            a = unyt_array([.3, .4, .5], 'cm')
            b = unyt_array([.1, .2, .3], 'dimensionless')
            c = np.array(b)
            d = unyt_array([1., 2., 3.], 'g')
            with pytest.raises(UnitOperationError):
                ufunc(a, b)
            with pytest.raises(UnitOperationError):
                ufunc(a, c)
            with pytest.raises(UnitOperationError):
                ufunc(a, d)
            binary_ufunc_comparison(ufunc, np.array(2.0), b)
            continue

        a = unyt_array([.3, .4, .5], 'cm')
        b = unyt_array([.1, .2, .3], 'cm')
        c = unyt_array([.1, .2, .3], 'm')
        d = unyt_array([.1, .2, .3], 'g')
        e = unyt_array([.1, .2, .3], 'erg/m**3')

        for pair in itertools.product([a, b, c, d, e], repeat=2):
            binary_ufunc_comparison(ufunc, pair[0], pair[1])


def test_reductions():
    arr = unyt_array([[1, 2, 3], [4, 5, 6]], 'cm')

    ev_result = arr.dot(unyt_array([1, 2, 3], 'cm'))
    res = unyt_array([14., 32.], 'cm**2')
    assert_equal(ev_result, res)
    assert_equal(ev_result.units, res.units)
    assert_isinstance(ev_result, unyt_array)

    qv = unyt_array([1, 2, 3], 'cm').dot(unyt_array([1, 2, 3], 'cm'))
    qa = unyt_quantity(14, 'cm**2')
    assert qv == qa
    assert qv.units == qa.units
    assert_isinstance(qv, unyt_quantity)

    answers = {
        'prod': (unyt_quantity(720, 'cm**6'),
                 unyt_array([4, 10, 18], 'cm**2'),
                 unyt_array([6, 120], 'cm**3')),
        'sum': (unyt_quantity(21, 'cm'),
                unyt_array([5., 7., 9.], 'cm'),
                unyt_array([6, 15], 'cm'),),
        'mean': (unyt_quantity(3.5, 'cm'),
                 unyt_array([2.5, 3.5, 4.5], 'cm'),
                 unyt_array([2, 5], 'cm')),
        'std': (unyt_quantity(1.707825127659933, 'cm'),
                unyt_array([1.5, 1.5, 1.5], 'cm'),
                unyt_array([0.81649658, 0.81649658], 'cm')),
    }
    for op, (result1, result2, result3) in answers.items():
        ev_result = getattr(arr, op)()
        assert_almost_equal(ev_result, result1)
        assert_equal(ev_result.units, result1.units)
        assert_isinstance(ev_result, unyt_quantity)
        for axis, result in [(0, result2), (1, result3), (-1, result3)]:
            ev_result = getattr(arr, op)(axis=axis)
            assert_almost_equal(ev_result, result)
            assert_equal(ev_result.units, result.units)
            assert_isinstance(ev_result, unyt_array)


def test_convenience():

    for orig in [[1., 2., 3.], (1., 2., 3.), np.array([1., 2., 3.]),
                 [[1.], [2.], [3.]], np.array([[1.], [2.], [3.]]),
                 [[1., 2., 3.]], np.array([[1., 2., 3.]])]:
        arr = unyt_array(orig, 'cm')
        arrou = unyt_array(orig, '1/cm')
        uoarr = unyt_array(1./np.array(orig), 'cm')

        assert_equal(arr.unit_quantity, unyt_quantity(1, 'cm'))
        assert_equal(arr.uq, unyt_quantity(1, 'cm'))
        assert_isinstance(arr.unit_quantity, unyt_quantity)
        assert_isinstance(arr.uq, unyt_quantity)

        assert_array_equal(arr.unit_array, unyt_array(np.ones_like(arr), 'cm'))
        assert_array_equal(arr.ua, unyt_array(np.ones_like(arr), 'cm'))
        assert_isinstance(arr.unit_array, unyt_array)
        assert_isinstance(arr.ua, unyt_array)

        for u in [arr.units, arr.unit_quantity, arr.unit_array, arr.uq,
                  arr.ua]:
            assert_array_equal(u*orig, arr)
            assert_array_equal(orig*u, arr)
            assert_array_equal(orig/u, arrou)
            assert_array_equal(u/orig, uoarr)

        assert_array_equal(arr.ndview, arr.view(np.ndarray))
        assert_array_equal(arr.d, arr.view(np.ndarray))
        assert arr.ndview.base is arr.base
        assert arr.d.base is arr.base

        assert_array_equal(arr.value, np.array(arr))
        assert_array_equal(arr.v, np.array(arr))


def test_registry_association():
    reg = UnitRegistry()
    a = unyt_quantity(3, 'cm', registry=reg)
    b = unyt_quantity(4, 'm')
    c = unyt_quantity(6, '', registry=reg)
    d = 5

    assert_equal(id(a.units.registry), id(reg))

    def binary_op_registry_comparison(op):
        e = op(a, b)
        f = op(b, a)
        g = op(c, d)
        h = op(d, c)

        assert_equal(id(e.units.registry), id(reg))
        assert_equal(id(f.units.registry), id(b.units.registry))
        assert_equal(id(g.units.registry), id(h.units.registry))
        assert_equal(id(g.units.registry), id(reg))

    def unary_op_registry_comparison(op):
        c = op(a)
        d = op(b)

        assert_equal(id(c.units.registry), id(reg))
        assert_equal(id(d.units.registry), id(b.units.registry))

    binary_ops = [operator.add, operator.sub, operator.mul,
                  operator.truediv]
    if hasattr(operator, "div"):
        binary_ops.append(operator.div)
    for op in binary_ops:
        binary_op_registry_comparison(op)

    for op in [operator.abs, operator.neg, operator.pos]:
        unary_op_registry_comparison(op)


def test_to_value():

    a = unyt_array([1.0, 2.0, 3.0], "kpc")
    assert_equal(a.to_value(), np.array([1.0, 2.0, 3.0]))
    assert_equal(a.to_value(), a.value)
    assert_equal(a.to_value("km"), a.in_units("km").value)

    b = unyt_quantity(5.5, "Msun")
    assert_equal(b.to_value(), 5.5)
    assert_equal(b.to_value("g"), b.in_units("g").value)


def test_astropy():
    if isinstance(_astropy.__version__, NotAModule):
        return
    ap_arr = np.arange(10)*_astropy.units.km/_astropy.units.hr
    yt_arr = unyt_array(np.arange(10), "km/hr")
    yt_arr2 = unyt_array.from_astropy(ap_arr)

    ap_quan = 10.*_astropy.units.Msun**0.5/(_astropy.units.kpc**3)
    yt_quan = unyt_quantity(10., "sqrt(Msun)/kpc**3")
    yt_quan2 = unyt_quantity.from_astropy(ap_quan)

    assert_array_equal(ap_arr, yt_arr.to_astropy())
    assert_array_equal(yt_arr, unyt_array.from_astropy(ap_arr))
    assert_array_equal(yt_arr, yt_arr2)

    assert_equal(ap_quan, yt_quan.to_astropy())
    assert_equal(yt_quan, unyt_quantity.from_astropy(ap_quan))
    assert_equal(yt_quan, yt_quan2)

    assert_array_equal(yt_arr, unyt_array.from_astropy(yt_arr.to_astropy()))
    assert_equal(yt_quan, unyt_quantity.from_astropy(yt_quan.to_astropy()))


def test_pint():
    if isinstance(_pint.UnitRegistry, NotAModule):
        return
    ureg = _pint.UnitRegistry()

    p_arr = np.arange(10)*ureg.km/ureg.year
    yt_arr = unyt_array(np.arange(10), "km/yr")
    yt_arr2 = unyt_array.from_pint(p_arr)

    p_quan = 10.*ureg.g**0.5/(ureg.mm**3)
    yt_quan = unyt_quantity(10., "sqrt(g)/mm**3")
    yt_quan2 = unyt_quantity.from_pint(p_quan)

    assert_array_equal(p_arr, yt_arr.to_pint())
    assert_equal(p_quan, yt_quan.to_pint())
    assert_array_equal(yt_arr, unyt_array.from_pint(p_arr))
    assert_array_equal(yt_arr, yt_arr2)

    assert_equal(p_quan.magnitude, yt_quan.to_pint().magnitude)
    assert_equal(p_quan, yt_quan.to_pint())
    assert_equal(yt_quan, unyt_quantity.from_pint(p_quan))
    assert_equal(yt_quan, yt_quan2)

    assert_array_equal(yt_arr, unyt_array.from_pint(yt_arr.to_pint()))
    assert_equal(yt_quan, unyt_quantity.from_pint(yt_quan.to_pint()))


def test_subclass():

    class unyt_a_subclass(unyt_array):
        pass

    a = unyt_a_subclass([4, 5, 6], 'g')
    b = unyt_a_subclass([7, 8, 9], 'kg')
    nu = unyt_a_subclass([10, 11, 12], '')
    nda = np.array([3, 4, 5])
    yta = unyt_array([6, 7, 8], 'mg')
    loq = [unyt_quantity(6, 'mg'), unyt_quantity(7, 'mg'),
           unyt_quantity(8, 'mg')]
    ytq = unyt_quantity(4, 'cm')
    ndf = np.float64(3)

    def op_comparison(op, inst1, inst2, compare_class):
        assert_isinstance(op(inst1, inst2), compare_class)
        assert_isinstance(op(inst2, inst1), compare_class)

    ops = [operator.mul, operator.truediv]
    if hasattr(operator, "div"):
        ops.append(operator.div)
    for op in ops:
        for inst in (b, ytq, ndf, yta, nda, loq):
            op_comparison(op, a, inst, unyt_a_subclass)

        op_comparison(op, ytq, nda, unyt_array)
        op_comparison(op, ytq, yta, unyt_array)

    for op in (operator.add, operator.sub):
        op_comparison(op, nu, nda, unyt_a_subclass)
        op_comparison(op, a, b, unyt_a_subclass)
        op_comparison(op, a, yta, unyt_a_subclass)
        op_comparison(op, a, loq, unyt_a_subclass)

    assert_isinstance(a[0], unyt_quantity)
    assert_isinstance(a[:], unyt_a_subclass)
    assert_isinstance(a[:2], unyt_a_subclass)
    assert_isinstance(unyt_a_subclass(yta), unyt_a_subclass)

    with pytest.raises(RuntimeError):
        a + 'hello'


def test_h5_io():
    if isinstance(_h5py.__version__, NotAModule):
        return

    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    reg = UnitRegistry()

    reg.add('code_length', 10.0, dimensions.length)

    warr = unyt_array(np.random.random((256, 256)), 'code_length',
                      registry=reg)

    warr.write_hdf5('test.h5')

    iarr = unyt_array.from_hdf5('test.h5')

    assert_equal(warr, iarr)
    assert_equal(warr.units.registry['code_length'],
                 iarr.units.registry['code_length'])

    # test code to overwrite existing dataset

    warr.write_hdf5('test.h5')

    giarr = unyt_array.from_hdf5('test.h5')

    assert_equal(warr, giarr)

    # test code to overwrite existing dataset with data that has a different
    # shape

    warr = unyt_array(np.random.random((255, 255)), 'code_length',
                      registry=reg)

    warr.write_hdf5('test.h5')

    giarr = unyt_array.from_hdf5('test.h5')

    assert_equal(warr, giarr)

    os.remove('test.h5')

    # write to a group that doesn't exist

    warr.write_hdf5('test.h5', dataset_name="test_dset",
                    group_name='/arrays/test_group')

    giarr = unyt_array.from_hdf5('test.h5', dataset_name="test_dset",
                                 group_name='/arrays/test_group')

    assert_equal(warr, giarr)

    os.remove('test.h5')

    # write to a group that does exist

    with _h5py.File('test.h5') as f:
        f.create_group('/arrays/test_group')

    warr.write_hdf5('test.h5', dataset_name="test_dset",
                    group_name='/arrays/test_group')

    giarr = unyt_array.from_hdf5('test.h5', dataset_name="test_dset",
                                 group_name='/arrays/test_group')

    assert_equal(warr, giarr)

    os.remove('test.h5')

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


def test_equivalencies():
    import unyt as u

    # equivalence is ignored if the conversion doesn't need one
    data = 12*u.g
    data.convert_to_equivalent('kg', None)
    assert data.value == .012
    assert data.units == u.kg

    data = 12*u.g
    data = data.to_equivalent('kg', None)
    assert data.value == .012
    assert data.units == u.kg

    # incorrect usage of an equivalence raises errors

    with pytest.raises(InvalidUnitEquivalence):
        data.convert_to_equivalent('erg', 'thermal')
    with pytest.raises(InvalidUnitEquivalence):
        data.convert_to_equivalent('m', 'mass_energy')
    with pytest.raises(InvalidUnitEquivalence):
        data.to_equivalent('erg', 'thermal')
    with pytest.raises(InvalidUnitEquivalence):
        data.to_equivalent('m', 'mass_energy')

    # Mass-energy

    mp = u.mp.copy()
    mp.convert_to_units("keV", "mass_energy")
    assert_allclose_units(u.mp.in_units("keV", "mass_energy"), mp)
    assert_allclose_units(mp, u.mp*u.clight*u.clight)
    assert_allclose_units(u.mp, mp.in_units("g", "mass_energy"))
    mp.convert_to_units("g", "mass_energy")
    assert_allclose_units(u.mp, mp)

    # Thermal

    T = 1e8*u.K
    E = T.in_units("W*hr", "thermal")
    assert_allclose_units(E, (u.kboltz*T).in_units("W*hr"))
    assert_allclose_units(T, E.in_units("K", "thermal"))

    T.convert_to_units("W*hr", "thermal")
    assert_allclose_units(E, T)
    T.convert_to_units("K", "thermal")
    assert_allclose_units(T, 1e8*u.K)

    # Spectral

    # wavelength to frequency

    lam = 4000*u.angstrom
    nu = lam.in_units("Hz", "spectral")
    assert_allclose_units(nu, u.clight/lam)
    lam.convert_to_units("MHz", "spectral")
    assert_allclose_units(lam, nu)
    assert lam.units == u.MHz.units
    assert nu.units == u.Hz.units

    # wavelength to photon energy

    lam = 4000*u.angstrom
    hnu = lam.in_units('erg', 'spectral')
    assert_allclose_units(hnu, u.hmks*u.clight/lam)
    lam.convert_to_units('eV', 'spectral')
    assert_allclose_units(lam, hnu)
    assert lam.units == u.eV.units
    assert hnu.units == u.erg.units

    # wavelength to spatial frequency

    lam = 4000*u.angstrom
    nubar = lam.in_units('1/angstrom', 'spectral')
    assert_allclose_units(nubar, 1/lam)
    lam.convert_to_units('1/cm', 'spectral')
    assert_allclose_units(lam, nubar)
    assert lam.units == (1 / u.cm).units
    assert nubar.units == (1 / u.angstrom).units

    # frequency to wavelength

    nu = 1*u.MHz
    lam = nu.to('km', 'spectral')
    assert_allclose_units(lam, u.clight/nu)
    nu.convert_to_units('m', 'spectral')
    assert_allclose_units(lam, nu)
    assert lam.units == u.km.units
    assert nu.units == u.m.units

    # frequency to spatial frequency

    nu = 1*u.MHz
    nubar = nu.to('1/km', 'spectral')
    assert_allclose_units(nubar, nu/u.clight)
    nu.convert_to_units('1/m', 'spectral')
    assert_allclose_units(nubar, nu)
    assert nubar.units == (1 / u.km).units
    assert nu.units == (1 / u.m).units

    # frequency to photon energy

    nu = 1*u.MHz
    E = nu.to('erg', 'spectral')
    assert_allclose_units(E, u.hmks*nu)
    nu.convert_to_units('J', 'spectral')
    assert_allclose_units(nu, E)
    assert nu.units == u.J.units
    assert E.units == u.erg.units

    # photon energy to frequency

    E = 13.6*u.eV
    nu = E.to('Hz', 'spectral')
    assert_allclose_units(nu, E/u.hmks)
    E.convert_to_units('MHz', 'spectral')
    assert_allclose_units(nu, E)
    assert E.units == u.MHz.units
    assert nu.units == u.Hz.units

    # photon energy to wavelength

    E = 13.6*u.eV
    lam = E.to('nm', 'spectral')
    assert_allclose_units(lam, u.hmks*u.clight/E)
    E.convert_to_units('angstrom', 'spectral')
    assert_allclose_units(E, lam)
    assert E.units == u.angstrom.units
    assert lam.units == u.nm.units

    # photon energy to spatial frequency

    E = 13.6*u.eV
    nubar = E.to('1/nm', 'spectral')
    assert_allclose_units(nubar, E/(u.hmks*u.clight))
    E.convert_to_units('1/angstrom', 'spectral')
    assert_allclose_units(E, nubar)
    assert E.units == (1 / u.angstrom).units
    assert nubar.units == (1 / u.nm).units

    # spatial frequency to frequency

    nubar = 1500. / u.cm
    nu = nubar.to('Hz', 'spectral')
    assert_allclose_units(nu, nubar*u.clight)
    nubar.convert_to_units('MHz', 'spectral')
    assert_allclose_units(nu, nubar)
    assert nubar.units == u.MHz.units
    assert nu.units == u.Hz.units

    # spatial frequency to wavelength

    nubar = 1500. / u.cm
    lam = nubar.to('nm', 'spectral')
    assert_allclose_units(lam, 1/nubar)
    nubar.convert_to_units('angstrom', 'spectral')
    assert_allclose_units(nubar, lam)
    assert nubar.units == u.angstrom.units
    assert lam.units == u.nm.units

    # spatial frequency to photon energy

    nubar = 1500. / u.cm
    E = nubar.to('erg', 'spectral')
    assert_allclose_units(E, u.hmks*u.clight*nubar)
    nubar.convert_to_units('J', 'spectral')
    assert_allclose_units(nubar, E)
    assert nubar.units == u.J.units
    assert E.units == u.erg.units

    # Sound-speed

    # tempearature <-> velocity

    mu = 0.6
    gg = 5./3.
    T = 1e8*u.K
    c_s = T.in_units("km/s", equivalence="sound_speed")
    assert_allclose_units(c_s, np.sqrt(gg*u.kboltz*T/(mu*u.mh)))
    assert_allclose_units(T, c_s.in_units("K", "sound_speed"))
    T.convert_to_units('m/s', 'sound_speed')
    assert_allclose_units(c_s, T)
    assert T.units == u.m.units/u.s.units
    assert c_s.units == u.km.units/u.s.units

    mu = 0.5
    gg = 4./3.
    T = 1e8*u.K
    c_s = T.in_units("km/s", "sound_speed", mu=mu, gamma=gg)
    assert_allclose_units(c_s, np.sqrt(gg*u.kboltz*T/(mu*u.mh)))
    assert_allclose_units(T, c_s.in_units("K", "sound_speed", mu=mu, gamma=gg))
    T.convert_to_units('m/s', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(c_s, T)
    assert T.units == u.m.units/u.s.units
    assert c_s.units == u.km.units/u.s.units

    # tempearture <-> energy

    mu = 0.5
    gg = 4./3.
    T = 1e8*u.K
    kT = T.in_units('eV', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(kT, u.kboltz*T)
    T.convert_to_units('erg', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(T, kT)
    assert T.units == u.erg.units
    assert kT.units == u.eV.units
    assert_allclose_units(
        T.in_units('K', 'sound_speed', mu=mu, gamma=gg), 1e8*u.K)
    kT.convert_to_units('K', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(kT, 1e8*u.K)

    # velocity <-> energy

    c_s = 300*u.m/u.s
    kT = c_s.in_units('erg', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(kT, c_s**2*mu*u.mh/gg)
    c_s.convert_to_units('J', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(c_s, kT)
    assert c_s.units == u.J.units
    assert kT.units == u.erg.units
    assert_allclose_units(
        kT.in_units('m/s', 'sound_speed', mu=mu, gamma=gg), 300*u.m/u.s)
    c_s.convert_to_units('m/s', 'sound_speed', mu=mu, gamma=gg)
    assert_allclose_units(c_s, 300*u.m/u.s)

    # Lorentz

    v = 0.8*u.clight
    g = v.in_units("dimensionless", "lorentz")
    g2 = unyt_quantity(1./np.sqrt(1.-0.8*0.8), "dimensionless")
    assert_allclose_units(g, g2)
    v.convert_to_units('', 'lorentz')
    assert_allclose_units(v, g2)
    v.convert_to_units('c', 'lorentz')
    v2 = g2.in_units("mile/hr", "lorentz")
    assert_allclose_units(v2, v.in_units("mile/hr"))

    # Schwarzschild

    msun = 1*u.Msun
    msun.convert_to_equivalent('km', 'schwarzschild')
    R = u.mass_sun_mks.in_units("kpc", "schwarzschild")
    assert_allclose_units(msun, R)
    assert_allclose_units(R.in_mks(), 2*u.G*u.mass_sun_mks/(u.clight**2))
    assert_allclose_units(u.mass_sun_mks, R.in_units("kg", "schwarzschild"))
    R.convert_to_units('Msun', 'schwarzschild')
    assert_allclose_units(u.mass_sun_mks, R)
    assert R.units == u.Msun.units
    assert msun.units == u.km.units

    # Compton

    me = 1*u.me
    me.convert_to_units('nm', 'compton')
    length = u.me.in_units("angstrom", "compton")
    assert_allclose_units(length, me)
    assert_allclose_units(length, u.hmks/(u.me*u.clight))
    assert_allclose_units(u.me, length.in_units("g", "compton"))
    assert me.units == u.nm.units
    assert length.units == u.angstrom.units
    me.convert_to_units('me', 'compton')
    assert_almost_equal(me.value, 1.0)

    # Number density

    rho = u.mp/u.m**3
    n = rho.in_units("m**-3", "number_density")
    assert_allclose_units(n, rho/(u.mh*0.6))
    assert_allclose_units(rho, n.in_units("kg/m**3", "number_density"))
    rho.convert_to_units('cm**-3', 'number_density')
    assert rho.units == (1/u.cm**3).units
    assert n.units == (1/u.m**3).units
    assert_allclose_units(n, rho)
    rho.convert_to_units('kg/m**3', 'number_density')
    assert_allclose_units(u.mp/u.m**3, rho)
    assert rho.units == (u.kg/u.m**3).units

    rho = u.mp/u.m**3
    n = rho.in_units("m**-3", equivalence="number_density", mu=0.75)
    assert_allclose_units(n, rho/(u.mh*0.75))
    assert_allclose_units(
        rho, n.in_units("kg/m**3", equivalence="number_density", mu=0.75))
    rho.convert_to_units('cm**-3', 'number_density', mu=0.75)
    assert rho.units == (1/u.cm**3).units
    assert n.units == (1/u.m**3).units
    assert_allclose_units(n, rho)
    rho.convert_to_units('kg/m**3', 'number_density', mu=0.75)
    assert_allclose_units(u.mp/u.m**3, rho)
    assert rho.units == (u.kg/u.m**3).units

    # Effective temperature

    T = 1e4*u.K
    F = T.in_units("W/m**2", equivalence="effective_temperature")
    assert_allclose_units(F, u.stefan_boltzmann_constant*T**4)
    assert_allclose_units(
        T, F.in_units("K", equivalence="effective_temperature"))
    T.convert_to_units('erg/s/cm**2', 'effective_temperature')
    assert_allclose_units(T, F)
    assert T.units == unyt_quantity(1.0, "erg/cm**2/s").units
    assert F.units == u.W/u.m**2
    T.convert_to_units('K', 'effective_temperature')
    assert_almost_equal(T.value, 1e4)
    assert T.units == u.K.units

    # to_value test

    assert_allclose_units(
        F.value,
        T.to_value("W/m**2", equivalence="effective_temperature"))
    assert_allclose_units(
        n.value,
        rho.to_value("m**-3", equivalence="number_density", mu=0.75))


def test_electromagnetic():
    import unyt as u

    # Various tests of SI and CGS electromagnetic units

    t = 1*u.Tesla
    g = 1*u.gauss
    assert t.to('gauss') == 1e4*u.gauss
    assert g.to('T') == 1e-4*u.Tesla

    qp_cgs = u.qp_cgs.in_units("C")
    assert_equal(qp_cgs.units.dimensions, dimensions.charge_mks)
    assert_almost_equal(qp_cgs.v, 10.0*u.qp.v/speed_of_light_cm_per_s)
    qp = 1*u.qp_cgs
    qp.convert_to_units("C")
    assert_equal(qp.units.dimensions, dimensions.charge_mks)
    assert_almost_equal(qp.v, 10*u.qp.v/u.clight.v)

    qp_cgs = u.qp.in_units("esu")
    assert_array_almost_equal(qp_cgs, u.qp_cgs)
    assert_equal(qp_cgs.units.dimensions, u.qp_cgs.units.dimensions)
    qp = u.qp.copy()
    qp.convert_to_units('esu')
    assert_almost_equal(qp_cgs, qp_cgs)
    assert qp.units == u.esu.units
    qp.convert_to_units('C')
    assert_almost_equal(u.qp, qp)
    assert qp.units == u.C.units

    qp_mks_k = u.qp_cgs.in_units("kC")
    assert_array_almost_equal(
        qp_mks_k.v, 1.0e-2*u.qp_cgs.v/speed_of_light_cm_per_s)
    qp = 1*u.qp_cgs
    qp.convert_to_units('kC')
    assert_almost_equal(qp, qp_mks_k)

    B = 1.0*u.T
    B_cgs = B.in_units("gauss")
    assert_equal(B.units.dimensions, dimensions.magnetic_field_mks)
    assert_equal(B_cgs.units.dimensions, dimensions.magnetic_field_cgs)
    assert_array_almost_equal(B_cgs, unyt_quantity(1.0e4, "gauss"))
    B_cgs = B.in_cgs()
    assert_equal(B.units.dimensions, dimensions.magnetic_field_mks)
    assert_equal(B_cgs.units.dimensions, dimensions.magnetic_field_cgs)
    assert_array_almost_equal(B_cgs, unyt_quantity(1.0e4, "gauss"))
    B_cgs = B.in_base('cgs')
    assert_equal(B.units.dimensions, dimensions.magnetic_field_mks)
    assert_equal(B_cgs.units.dimensions, dimensions.magnetic_field_cgs)
    assert_array_almost_equal(B_cgs, unyt_quantity(1.0e4, "gauss"))
    B.convert_to_cgs()
    assert_almost_equal(B, B_cgs)
    B.convert_to_mks()
    B_cgs2 = B.to("gauss")
    assert_almost_equal(B_cgs, B_cgs2)
    B_mks2 = B_cgs2.to("T")
    assert_almost_equal(B, B_mks2)

    B = 1.0*u.T
    u_mks = B*B/(2*u.mu_0)
    assert_equal(u_mks.units.dimensions, dimensions.pressure)
    u_cgs = B_cgs*B_cgs/(8*np.pi)
    with pytest.raises(UnitConversionError):
        u_cgs.to(u_mks.units)
    with pytest.raises(UnitConversionError):
        u_mks.to(u_cgs.units)
    with pytest.raises(UnitsNotReducible):
        u_mks.in_cgs()
    with pytest.raises(UnitsNotReducible):
        u_cgs.in_mks()

    current = 1.0*u.A
    I_cgs = current.in_units("statA")
    assert_array_almost_equal(
        I_cgs, unyt_quantity(0.1*speed_of_light_cm_per_s, "statA"))
    assert_array_almost_equal(I_cgs.in_units("mA"), current.in_units("mA"))
    assert_equal(I_cgs.units.dimensions, dimensions.current_cgs)
    current.convert_to_units('statA')
    assert current.units == u.statA.units
    current.convert_to_units('A')
    assert current.units == u.A.units
    I_cgs2 = current.to('statA')
    assert I_cgs2.units == u.statA.units
    assert_array_almost_equal(
        I_cgs2, unyt_quantity(0.1*speed_of_light_cm_per_s, "statA"))

    current = 1.0*u.A
    R = unyt_quantity(1.0, "ohm")
    R_cgs = R.in_units("statohm")
    P_mks = current*current*R
    P_cgs = I_cgs*I_cgs*R_cgs
    assert_equal(P_mks.units.dimensions, dimensions.power)
    assert_equal(P_cgs.units.dimensions, dimensions.power)
    with pytest.raises(UnitsNotReducible):
        P_cgs.in_cgs()
    with pytest.raises(UnitsNotReducible):
        P_mks.in_cgs()

    V = unyt_quantity(1.0, "statV")
    V_mks = V.in_units("V")
    assert_array_almost_equal(V_mks.v, 1.0e8*V.v/speed_of_light_cm_per_s)

    data = 1.0*u.C*u.T*u.V
    with pytest.raises(UnitConversionError):
        data.to('statC*G*statV')
    with pytest.raises(UnitConversionError):
        data.convert_to_units('statC*G*statV')
    with pytest.raises(UnitsNotReducible):
        data.in_cgs()

    data = 1.0*u.statC*u.G*u.statV
    with pytest.raises(UnitConversionError):
        data.to('C*T*V')
    with pytest.raises(UnitConversionError):
        data.convert_to_units('C*T*V')
    with pytest.raises(UnitsNotReducible):
        data.in_mks()

    mu_0 = 4.0e-7*math.pi*u.N/u.A**2
    eps_0 = 8.85418781782e-12*u.m**-3/u.kg*u.s**4*u.A**2
    assert_almost_equal((1.0/(u.clight**2*mu_0)).in_units(eps_0.units), eps_0)


def test_ytarray_coercion():
    a = unyt_array([1, 2, 3], 'cm')
    q = unyt_quantity(3, 'cm')
    na = np.array([1, 2, 3])

    assert_isinstance(a*q, unyt_array)
    assert_isinstance(q*na, unyt_array)
    assert_isinstance(q*3, unyt_quantity)
    assert_isinstance(q*np.float64(3), unyt_quantity)
    assert_isinstance(q*np.array(3), unyt_quantity)


def test_numpy_wrappers():
    a1 = unyt_array([1, 2, 3], 'cm')
    a2 = unyt_array([2, 3, 4, 5, 6], 'cm')
    a3 = unyt_array([[1, 2, 3], [4, 5, 6]], 'cm')
    catenate_answer = [1, 2, 3, 2, 3, 4, 5, 6]
    intersect_answer = [2, 3]
    union_answer = [1, 2, 3, 4, 5, 6]
    cross_answer = [-2, 4, -2]
    norm_answer = np.sqrt(1**2 + 2**2 + 3**2)
    arr_norm_answer = [norm_answer, np.sqrt(4**2 + 5**2 + 6**2)]
    dot_answer = 14

    assert_array_equal(unyt_array(catenate_answer, 'cm'),
                       uconcatenate((a1, a2)))
    assert_array_equal(catenate_answer, np.concatenate((a1, a2)))

    assert_array_equal(unyt_array(intersect_answer, 'cm'),
                       uintersect1d(a1, a2))
    assert_array_equal(intersect_answer, np.intersect1d(a1, a2))

    assert_array_equal(unyt_array(union_answer, 'cm'), uunion1d(a1, a2))
    assert_array_equal(union_answer, np.union1d(a1, a2))

    assert_array_equal(unyt_array(cross_answer, 'cm**2'),
                       ucross(a1, a1+(2*a1.units)))
    assert_array_equal(cross_answer, np.cross(a1.v, a1.v+2))

    assert_array_equal(unorm(a1), unyt_quantity(norm_answer, 'cm'))
    assert_array_equal(np.linalg.norm(a1), norm_answer)
    assert_array_equal(unorm(a3, axis=1), unyt_array(arr_norm_answer, 'cm'))
    assert_array_equal(np.linalg.norm(a3, axis=1), arr_norm_answer)

    assert_array_equal(udot(a1, a1), unyt_quantity(dot_answer, 'cm**2'))

    assert_array_equal(np.array(catenate_answer), uconcatenate((a1.v, a2.v)))
    with pytest.raises(RuntimeError):
        uconcatenate((a1, a2.v))
    with pytest.raises(RuntimeError):
        uconcatenate((a1.to('m'), a2))


def test_dimensionless_conversion():
    a = unyt_quantity(1, 'Zsun')
    b = a.in_units('Zsun')
    a.convert_to_units('Zsun')
    assert a.units.base_value == metallicity_sun
    assert b.units.base_value == metallicity_sun


def test_modified_unit_division():
    reg1 = UnitRegistry()
    reg2 = UnitRegistry()

    reg1.modify('g', 50)

    a = unyt_quantity(3, 'g', registry=reg1)
    b = unyt_quantity(3, 'g', registry=reg2)

    ret = a/b
    assert ret == 50000.
    assert ret.units.is_dimensionless
    assert ret.units.base_value == 1.0


def test_loadtxt_and_savetxt():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    a = unyt_array(np.random.random(10), "kpc")
    b = unyt_array(np.random.random(10), "Msun")
    c = unyt_array(np.random.random(10), "km/s")

    savetxt("arrays.dat", [a, b, c], delimiter=",")

    d, e = loadtxt("arrays.dat", usecols=(1, 2), delimiter=",")

    assert_array_equal(b, d)
    assert_array_equal(c, e)

    # adding newlines to the file doesn't matter

    savetxt("arrays.dat", [a, b, c], delimiter=",")

    with open('arrays.dat', 'r+') as f:
        content = f.read()
        f.seek(0, 0)
        f.write('\n' + content)

    d, e = loadtxt("arrays.dat", usecols=(1, 2), delimiter=",")

    assert_array_equal(b, d)
    assert_array_equal(c, e)

    # data saved by numpy savetxt are loaded without units

    np.savetxt("arrays.dat", np.squeeze(np.transpose([a.v, b.v, c.v])),
               delimiter=",")

    d, e = loadtxt("arrays.dat", usecols=(1, 2), delimiter=",")

    assert_array_equal(b.v, d)
    assert_array_equal(c.v, e)

    # save a single array

    savetxt("arrays.dat", a)

    d = loadtxt("arrays.dat")

    assert_array_equal(a, d)

    # save an array with no units and an array with units with a header

    savetxt("arrays.dat", [a.v, b], header='this is a header!')

    d, e = loadtxt("arrays.dat")

    assert_array_equal(a.v, d)
    assert_array_equal(b, e)

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


def test_trig_ufunc_degrees():
    for ufunc in (np.sin, np.cos, np.tan):
        degree_values = np.random.random(10)*degree
        radian_values = degree_values.in_units('radian')
        assert_array_equal(ufunc(degree_values), ufunc(radian_values))


def test_builtin_sum():
    from unyt import km

    arr = [1, 2, 3]*km
    assert_equal(sum(arr), 6*km)


def test_initialization_different_registries():

    reg1 = UnitRegistry()
    reg2 = UnitRegistry()

    reg1.add('code_length', 1.0, dimensions.length)
    reg2.add('code_length', 3.0, dimensions.length)

    l1 = unyt_quantity(1.0, 'code_length', registry=reg1)
    l2 = unyt_quantity(1.0, 'code_length', registry=reg2)

    assert_almost_equal(float(l1.in_mks()), 1.0)
    assert_almost_equal(float(l2.in_mks()), 3.0)


def test_ones_and_zeros_like():
    data = unyt_array([1, 2, 3], 'cm')
    zd = np.zeros_like(data)
    od = np.ones_like(data)

    assert_equal(zd, unyt_array([0, 0, 0], 'cm'))
    assert_equal(zd.units, data.units)
    assert_equal(od, unyt_array([1, 1, 1], 'cm'))
    assert_equal(od.units, data.units)


def test_coerce_iterable():
    from unyt import cm, km

    a = unyt_array([1, 2, 3], 'cm')
    b = [1*cm, 2*km, 3*cm]

    with pytest.raises(IterableUnitCoercionError):
        a + b
    with pytest.raises(IterableUnitCoercionError):
        b + a
    with pytest.raises(IterableUnitCoercionError):
        unyt_array(b)


def test_bypass_validation():
    from unyt import unyt_array, cm, UnitRegistry

    obj = unyt_array(np.array([1., 2., 3.]), cm, bypass_validation=True)
    assert obj.units is cm

    reg = UnitRegistry()
    obj = unyt_array(np.array([1., 2., 3.]), cm, registry=reg,
                     bypass_validation=True)
    assert obj.units == cm
    assert obj.units.registry is reg


def test_creation():
    from unyt import cm, UnitRegistry

    data = [1, 2, 3]*cm

    new_data = unyt_array(data)

    assert new_data.units is cm
    assert_array_equal(new_data.v, np.array([1, 2, 3], dtype='float64'))

    reg = UnitRegistry()

    new_data = unyt_array(data, registry=reg)
    assert_array_equal(new_data.v, np.array([1, 2, 3], dtype='float64'))
    assert new_data.units is not cm
    assert new_data.units == cm
    assert new_data.units.registry is reg

    new_data = unyt_array([1, 2, 3], cm)
    assert_array_equal(new_data.v, np.array([1, 2, 3], dtype='float64'))
    assert new_data.units is cm

    new_data = unyt_array([1, 2, 3], cm, registry=reg)
    assert_array_equal(new_data.v, np.array([1, 2, 3], dtype='float64'))
    assert new_data.units is not cm
    assert new_data.units == cm
    assert new_data.units.registry is reg

    with pytest.raises(RuntimeError):
        unyt_quantity('hello', 'cm')
    with pytest.raises(RuntimeError):
        unyt_quantity(np.array([1, 2, 3]), 'cm')


def test_round():
    from unyt import km

    assert_equal(round(3.3*km), 3.)
    assert_equal(round(3.5*km), 4.)
    assert_equal(round(3*km), 3)
    assert_equal(round(3.7*km), 4)

    with pytest.raises(TypeError):
        round([1, 2, 3]*km)



"""
Predefined useful aliases to physical units

Note that all of these names can be imported from the top-level unyt namespace.
For example::

    >>> from unyt import cm, g, s
    >>> data = [3, 4, 5]*g*cm/s
    >>> print(data)
    [3. 4. 5.] cm*g/s

"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from unyt.unit_object import Unit

#
# meter
#

#: femtometer
fm = femtometer = Unit("fm")
#: picometer
pm = picometer = Unit("pm")
#: nanometer
nm = nanometer = Unit("nm")
#: micrometer
um = micrometer = Unit("um")
#: millimeter
mm = millimeter = Unit("mm")
#: centimeter
cm = centimeter = Unit("cm")
#: meter
m = meter = Unit("m")
#: kilometer
km = kilometer = Unit("km")
#: Megameter
Mm = Megameter = megameter = Unit("Mm")

#
# parsec
#

#: parsec
pc = parsec = Unit("pc")
#: kiloparsec
kpc = kiloparsec = Unit("kpc")
#: megaparsec
Mpc = mpc = megaparsec = Unit("Mpc")
#: gigaparsec
Gpc = gpc = Gigaparsec = Unit("Gpc")

#
# gram
#

#: picogram
pg = picogram = Unit("pg")
#: nanogram
ng = nanogram = Unit("ng")
#: micogram
ug = microgram = Unit("ug")
#: milligram
mg = milligram = Unit("mg")
#: gram
g = gram = Unit("g")
#: kilogram
kg = kilogram = Unit("kg")
#: megagram
Mg = megagramme = tonne = metric_ton = Unit("Mg")

#
# second
#

#: femtosecond
fs = femtoseconds = Unit("fs")
#: picosecond
ps = picosecond = Unit("ps")
#: nanosecond
ns = nanosecond = Unit("ns")
#: millisecond
ms = millisecond = Unit("ms")
#: second
s = second = Unit("s")
#: kilosecond
ks = kilosecond = Unit("ks")
#: megasecond
Ms = megasecond = Unit("Ms")
#: gigasecond
Gs = gigasecond = Unit("Gs")

#
# minute
#

#: minute
min = minute = Unit("min")

#
# hr
#

#: hour
hr = hour = Unit("hr")

#
# day
#

#: day
day = Unit("day")

#
# year
#

#: year
yr = year = Unit("yr")
#: kiloyear
kyr = kiloyear = Unit("kyr")
#: Megayear
Myr = Megayear = megayear = Unit("Myr")
#: Gigayear
Gyr = Gigayear = gigayear = Unit("Gyr")

#
# Temperatures
#

#: Degree kelvin
degree_kelvin = Kelvin = K = Unit("K")
#: Degree fahrenheit
degree_fahrenheit = degF = Unit("degF")
#: Degree Celsius
degree_celsius = degC = Unit("degC")
#:
degree_rankine = R = Unit("R")

#
# Misc CGS
#

#: dyne (CGS force)
dyne = dyn = Unit("dyne")
#: erg (CGS energy)
erg = ergs = Unit("erg")
#:
barye = Ba = Unit("Ba")

#
# Misc SI
#

#: Newton (SI force)
N = Newton = newton = Unit("N")
#: Joule (SI energy)
J = Joule = joule = Unit("J")
#: Watt (SI power)
W = Watt = watt = Unit("W")
#: Hertz
Hz = Hertz = hertz = Unit("Hz")
#: Kilohertz
kHz = khz = kilohertz = Unit("kHz")
#: Megahertz
MHz = mhz = megahertz = Unit("MHz")
#: Gigahertz
GHz = ghz = gigahertz = Unit("GHz")
#: Terahertz
THz = thz = terahertz = Unit("THz")
#: Pascal
Pa = pascal = Unit("Pa")
#: sievert
Sv = sievert = Unit("Sv")
#: candela
cd = Candela = candela = Unit("cd")
#: lumen
lm = Lumen = lumen = Unit("lm")
#: lux
lx = Lux = lux = Unit('lx')
#: lambert
lambert = Lambert = Unit('lambert')
#: nit
nt = nit = Nit = Unit("nt")

#
# Imperial units
#

#: inch
inch = Unit("inch")
#: foot
ft = foot = Unit("ft")
#: mile
mile = Unit("mile")
#: furlong
fur = furlong = Unit("fur")
#: yard
yard = yd = Unit("yd")
#: pound
lb = pound = Unit("lb")
#: slug
slug = Unit("slug")
#: ounce
oz = ounce = Unit("oz")
#: ton
ton = Unit("ton")
#: pound-force
lbf = pound_force = Unit("lbf")
#: atmosphere
atm = atmosphere = Unit("atm")
#: horsepower
hp = horsepower = Unit("hp")
#: BTU
BTU = Unit("BTU")
#: calorie
cal = calorie = Unit("cal")
#: psi
psi = Unit("psi")

#
# Solar units
#

#: Mass of the sun
Msun = Unit("Msun")
#: Radius of the sun
Rsun = R_sun = solar_radius = Unit("Rsun")
#: Radius of the sun
rsun = r_sun = Unit("rsun")
#: Luminosity of the sun
Lsun = lsun = l_sun = solar_luminosity = Unit("Lsun")
#: Temperature of the sun
Tsun = T_sun = solar_temperature = Unit("Tsun")
#: Metallicity of the sun
Zsun = Z_sun = solar_metallicity = Unit("Zsun")

#
# Misc Astronomical units
#

#: Astronomical unit
AU = astronomical_unit = Unit("AU")
#: Astronomical unit
au = Unit("au")
#: Light year
ly = light_year = Unit("ly")
#: Radius of the Earth
Rearth = R_earth = earth_radius = Unit('R_earth')
#: Radius of the Earth
rearth = r_earth = Unit('r_earth')
#: Radius of Jupiter
Rjup = R_jup = jupiter_radius = Unit('R_jup')
#: Radius of Jupiter
rjup = r_jup = Unit('r_jup')
#: Jansky
Jy = jansky = Unit("Jy")

#
# Physical units
#

#: electronvolt
eV = electron_volt = electronvolt = Unit("eV")
#: kiloelectronvolt
keV = kilo_electron_volt = kiloelectronvolt = Unit("keV")
#: Megaelectronvolt
MeV = mega_electron_volt = Megaelectronvolt = megaelectronvolt = Unit("MeV")
#: Gigaelectronvolt
GeV = giga_electron_volt = Gigaelectronvolt = gigaelectronvolt = Unit("GeV")
#: Atomic mass unit
amu = atomic_mass_unit = Unit("amu")
mol = Unit("mol")
#: Angstrom
angstrom = Unit("angstrom")
#: Electron mass
me = electron_mass = Unit("me")

#
# Angle units
#

#: Degree (angle)
deg = degree = Unit("degree")
#: Radian
rad = radian = Unit("radian")
#: Arcsecond
arcsec = arcsecond = Unit("arcsec")
#: Arcminute
arcmin = arcminute = Unit("arcmin")
#: milliarcsecond
mas = milliarcsecond = Unit("mas")
#: steradian
sr = steradian = Unit("steradian")
#: hourangle
HA = hourangle = Unit("hourangle")


#
# CGS electromagnetic units
#

#: electrostatic unit
electrostatic_unit = esu = Unit("esu")
#: Gauss
gauss = G = Unit("gauss")
#: Statcoulomb
statcoulomb = statC = Unit("statC")
#: Statampere
statampere = statA = Unit("statA")
#: Statvolt
statvolt = statV = Unit("statV")
#: Statohm
statohm = Unit("statohm")
#: Maxwell
Mx = maxwell = Maxwell = Unit("Mx")

#
# SI electromagnetic units
#

#: Coulomb
C = coulomb = Coulomb = Unit("C")
#: Tesla
T = tesla = Tesla = Unit("T")
#: Ampere
A = ampere = Ampere = Unit("A")
#: Volt
V = volt = Volt = Unit("V")
#: Ohm
ohm = Ohm = Unit("ohm")
#: Weber
Wb = weber = Weber = Unit("Wb")

#
# Geographic units
#

#: Degree latitude
latitude = lat = Unit("lat")
#: Degree longitude
longitude = lon = Unit("lon")

#
# Misc dimensionless units
#

#: Count of something
counts = count = Unit('counts')
#: Number of photons
photons = photon = Unit('photons')
#: dimensionless
_ = dimensionless = Unit('')



"""
Unit system class.

"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

from collections import OrderedDict
from six import string_types
from unyt import dimensions
from unyt.exceptions import (
    MissingMKSCurrent,
    IllDefinedUnitSystem
)
from unyt.unit_object import (
    Unit,
    _get_system_unit_string
)

unit_system_registry = {}

cmks = dimensions.current_mks


class UnitSystem(object):
    """
    Create a UnitSystem for facilitating conversions to a default set of units.

    Parameters
    ----------
    name : string
        The name of the unit system. Will be used as the key in the
        *unit_system_registry* dict to reference the unit system by.
    length_unit : string
        The base length unit of this unit system.
    mass_unit : string
        The base mass unit of this unit system.
    time_unit : string
        The base time unit of this unit system.
    temperature_unit : string, optional
        The base temperature unit of this unit system. Defaults to "K".
    angle_unit : string, optional
        The base angle unit of this unit system. Defaults to "rad".
    mks_system: boolean, optional
        Whether or not this unit system has SI-specific units.
        Default: False
    current_mks_unit : string, optional
        The base current unit of this unit system. Defaults to "A".
    luminous_intensity_unit : string, optional
        The base luminous intensity unit of this unit system.
        Defaults to "cd".
    registry : :class:`yt.units.unit_registry.UnitRegistry` object
        The unit registry associated with this unit system. Only
        useful for defining unit systems based on code units.
    """
    def __init__(self, name, length_unit, mass_unit, time_unit,
                 temperature_unit="K", angle_unit="rad",
                 current_mks_unit="A", luminous_intensity_unit="cd",
                 registry=None):
        self.registry = registry
        if current_mks_unit is not None:
            current_mks_unit = Unit(current_mks_unit,
                                    registry=self.registry)
        self.units_map = OrderedDict([
            (dimensions.length, Unit(length_unit, registry=self.registry)),
            (dimensions.mass, Unit(mass_unit, registry=self.registry)),
            (dimensions.time, Unit(time_unit, registry=self.registry)),
            (dimensions.temperature, Unit(
                temperature_unit, registry=self.registry)),
            (dimensions.angle, Unit(angle_unit, registry=self.registry)),
            (dimensions.current_mks, current_mks_unit),
            (dimensions.luminous_intensity, Unit(
                luminous_intensity_unit, registry=self.registry))])
        for dimension, unit in self.units_map.items():
            # CGS sets the current_mks unit to none, so catch it here
            if unit is None:
                continue
            if unit.dimensions is not dimension:
                raise IllDefinedUnitSystem(self.units_map)
        self._dims = ["length", "mass", "time", "temperature", "angle",
                      "current_mks", "luminous_intensity"]
        self.registry = registry
        self.base_units = self.units_map.copy()
        unit_system_registry[name] = self
        self.name = name

    def __getitem__(self, key):
        if isinstance(key, string_types):
            key = getattr(dimensions, key)
        um = self.units_map
        if key not in um or um[key] is None or um[key].dimensions is not key:
            if cmks in key.free_symbols and self.units_map[cmks] is None:
                raise MissingMKSCurrent(self.name)
            units = _get_system_unit_string(key, self.units_map)
            self.units_map[key] = Unit(units, registry=self.registry)
        return self.units_map[key]

    def __setitem__(self, key, value):
        if isinstance(key, string_types):
            if key not in self._dims:
                self._dims.append(key)
            key = getattr(dimensions, key)
        if self.units_map[cmks] is None and cmks in key.free_symbols:
            raise MissingMKSCurrent(self.name)
        self.units_map[key] = Unit(value, registry=self.registry)

    def __str__(self):
        return self.name

    def __repr__(self):
        repr = "%s Unit System\n" % self.name
        repr += " Base Units:\n"
        for dim in self.base_units:
            if self.base_units[dim] is not None:
                repr += "  %s: %s\n" % (
                    str(dim).strip("()"), self.base_units[dim])
        repr += " Other Units:\n"
        for key in self._dims:
            dim = getattr(dimensions, key)
            if dim not in self.base_units:
                repr += "  %s: %s\n" % (key, self.units_map[dim])
        return repr[:-1]


#: The CGS unit system
cgs_unit_system = UnitSystem("cgs", "cm", "g", "s", current_mks_unit=None)
cgs_unit_system["energy"] = "erg"
cgs_unit_system["specific_energy"] = "erg/g"
cgs_unit_system["pressure"] = "dyne/cm**2"
cgs_unit_system["force"] = "dyne"
cgs_unit_system["magnetic_field_cgs"] = "gauss"
cgs_unit_system["charge_cgs"] = "esu"
cgs_unit_system["current_cgs"] = "statA"
cgs_unit_system["power"] = "erg/s"

#: The MKS unit system
mks_unit_system = UnitSystem("mks", "m", "kg", "s")
mks_unit_system["energy"] = "J"
mks_unit_system["specific_energy"] = "J/kg"
mks_unit_system["pressure"] = "Pa"
mks_unit_system["force"] = "N"
mks_unit_system["magnetic_field_mks"] = "T"
mks_unit_system["charge_mks"] = "C"
mks_unit_system["power"] = "W"

#: The imperial unit system
imperial_unit_system = UnitSystem("imperial", "ft", "lb", "s",
                                  temperature_unit="R")
imperial_unit_system["force"] = "lbf"
imperial_unit_system["energy"] = "ft*lbf"
imperial_unit_system["pressure"] = "lbf/ft**2"
imperial_unit_system["power"] = "hp"

#: The galactic unit system
galactic_unit_system = UnitSystem("galactic", "kpc", "Msun", "Myr")
galactic_unit_system["energy"] = "keV"
galactic_unit_system["magnetic_field_cgs"] = "uG"

#: The solar unit system
solar_unit_system = UnitSystem("solar", "AU", "Mearth", "yr")

#: Geometrized unit system
geometrized_unit_system = UnitSystem("geometrized", "l_geom",
                                     "m_geom", "t_geom")

#: Planck unit system
planck_unit_system = UnitSystem("planck", "l_pl", "m_pl", "t_pl",
                                temperature_unit="T_pl")
planck_unit_system["energy"] = "E_pl"
planck_unit_system["charge_mks"] = "q_pl"



"""
Test unit lookup tables and registry




"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import pytest

from unyt.dimensions import length
from unyt.exceptions import (
    SymbolNotFoundError,
    UnitParseError,
)
from unyt.unit_object import Unit
from unyt.unit_registry import UnitRegistry


def test_add_modify_error():
    from unyt import m
    ureg = UnitRegistry()

    with pytest.raises(UnitParseError):
        ureg.add('tayne', 1, length)
    with pytest.raises(UnitParseError):
        ureg.add('tayne', 'blah', length)
    with pytest.raises(UnitParseError):
        ureg.add('tayne', 1.0, length, offset=1)
    with pytest.raises(UnitParseError):
        ureg.add('tayne', 1.0, length, offset='blah')

    ureg.add('tayne', 1.1, length)

    with pytest.raises(SymbolNotFoundError):
        ureg.remove('tayn')
    with pytest.raises(SymbolNotFoundError):
        ureg.modify('tayn', 1.2)

    ureg.modify('tayne', 1.0*m)

    assert ureg['tayne'][:3] == ureg['m'][:3]


def test_keys():
    ureg = UnitRegistry()
    assert sorted(ureg.keys()) == sorted(ureg.lut.keys())


def test_prefixable_units():
    ureg = UnitRegistry()
    pu = ureg.prefixable_units
    assert 'm' in pu
    assert 'pc' in pu
    assert 'mol' in pu
    ureg.add('foobar', 1.0, length, prefixable=True)
    assert 'foobar' in ureg.prefixable_units
    mfoobar = Unit('mfoobar', registry=ureg)
    foobar = Unit('foobar', registry=ureg)
    assert (1*foobar)/(1*mfoobar) == 1000


def test_registry_contains():
    ureg = UnitRegistry()
    assert 'm' in ureg
    assert 'cm' in ureg
    assert 'erg' in ureg
    assert 'Merg' in ureg
    assert 'foobar' not in ureg
    assert Unit('m', registry=ureg) in ureg



"""
Exception classes defined by unyt



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------


class UnitOperationError(ValueError):
    """An exception that is raised when unit operations are not allowed

    Example
    -------

    >>> import unyt as u
    >>> 3*u.g + 4*u.m  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.UnitOperationError: The <ufunc 'add'> operator
    for unyt_arrays with units "g" (dimensions "(mass)") and
    "m" (dimensions "(length)") is not well defined.
    """
    def __init__(self, operation, unit1, unit2=None):
        self.operation = operation
        self.unit1 = unit1
        self.unit2 = unit2
        ValueError.__init__(self)

    def __str__(self):
        err = ("The %s operator for unyt_arrays with units \"%s\" "
               "(dimensions \"%s\") " %
               (self.operation, self.unit1, self.unit1.dimensions))
        if self.unit2 is not None:
            err += ("and \"%s\" (dimensions \"%s\") " %
                    (self.unit2, self.unit2.dimensions))
        err += "is not well defined."
        return err


class UnitConversionError(Exception):
    """An error raised when converting to a unit with different dimensions.

    Example
    -------

    >>> import unyt as u
    >>> data = 3*u.g
    >>> data.to('m')  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.UnitConversionError: Cannot convert between 'g'
    (dim '(mass)') and 'm' (dim '(length)').
    """
    def __init__(self, unit1, dimension1, unit2, dimension2):
        self.unit1 = unit1
        self.unit2 = unit2
        self.dimension1 = dimension1
        self.dimension2 = dimension2
        Exception.__init__(self)

    def __str__(self):
        err = ("Cannot convert between '%s' (dim '%s') and '%s' "
               "(dim '%s')." % (self.unit1, self.dimension1, self.unit2,
                                self.dimension2))
        return err


class MissingMKSCurrent(Exception):
    """Raised when querying a unit system for MKS current dimensions

    Since current is a base dimension for SI or SI-like unit systems but not in
    CGS or CGS-like unit systems, dimensions that include the MKS current
    dimension (the dimension of ampere) are not representable in CGS-like unit
    systems. When a CGS-like unit system is queried for such a dimension, this
    error is raised.

    Example
    -------

    >>> from unyt.unit_systems import cgs_unit_system as us
    >>> from unyt import ampere
    >>> us[ampere.dimensions]  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.MissingMKSCurrent: The cgs unit system does not
    have a MKS current base unit

    """
    def __init__(self, unit_system_name):
        self.unit_system_name = unit_system_name

    def __str__(self):
        err = ("The %s unit system does not have a MKS current base unit" %
               self.unit_system_name)
        return err


class MKSCGSConversionError(Exception):
    """Raised when conversion between MKS and CGS units cannot be performed

    This error is raised and caught internally and will expose itself
    to the user as part of a chained exception leading to a
    UnitConversionError.
    """
    def __init__(self, unit):
        self.unit = unit

    def __str__(self):
        err = ("The '%s' unit cannot be safely converted." % self.unit)
        return err


class UnitsNotReducible(Exception):
    """Raised when a unit cannot be safely represented in a unit system

    Example
    -------

    >>> from unyt import A, cm
    >>> data = 12*A/cm
    >>> data.in_cgs()  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.UnitsNotReducible: The unit "A/cm" (dimensions
    "(current_mks)/(length)") cannot be reduced to an expression
    within the cgs system of units.
    """
    def __init__(self, unit, units_base):
        self.unit = unit
        self.units_base = units_base
        Exception.__init__(self)

    def __str__(self):
        err = ("The unit \"%s\" (dimensions \"%s\") cannot be reduced to an "
               "expression within the %s system of units." %
               (self.unit, self.unit.dimensions, self.units_base))
        return err


class IterableUnitCoercionError(Exception):
    """Raised when an iterable cannot be converted to a unyt_array

    Example
    -------

    >>> from unyt import km, cm, unyt_array
    >>> data = [2*cm, 3*km]
    >>> unyt_array(data)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.IterableUnitCoercionError: Received a list or
    tuple of quantities with nonuniform units:
    [unyt_quantity(2., 'cm'), unyt_quantity(3., 'km')]
    """
    def __init__(self, quantity_list):
        self.quantity_list = quantity_list

    def __str__(self):
        err = ("Received a list or tuple of quantities with nonuniform units: "
               "%s" % self.quantity_list)
        return err


class InvalidUnitEquivalence(Exception):
    """Raised an equivalence does not apply to a unit conversion

    Example
    -------

    >>> import unyt as u
    >>> data = 12*u.g
    >>> data.to('erg', equivalence='thermal')\
 # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.InvalidUnitEquivalence: The unit equivalence
    'thermal' does not exist for the units 'g' and 'erg'.
    """
    def __init__(self, equiv, unit1, unit2):
        self.equiv = equiv
        self.unit1 = unit1
        self.unit2 = unit2

    def __str__(self):
        from unyt.unit_object import Unit
        if isinstance(self.unit2, Unit):
            msg = ("The unit equivalence '%s' does not exist for the units "
                   "'%s' and '%s'.")
        else:
            msg = ("The unit equivalence '%s' does not exist for units '%s' "
                   "to convert to a new unit with dimensions '%s'.")
        return msg % (self.equiv, self.unit1, self.unit2)


class InvalidUnitOperation(Exception):
    """Raised when an operation on a unit object is not allowed

    Example
    -------

    >>> from unyt import cm, g
    >>> cm + g  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.InvalidUnitOperation: addition with unit objects
    is not allowed
    """
    pass


class SymbolNotFoundError(Exception):
    """Raised when a unit name is not available in a unit registry

    Example
    -------

    >>> from unyt.unit_registry import default_unit_registry
    >>> default_unit_registry['made_up_unit']\
  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.SymbolNotFoundError: The symbol 'made_up_unit'
    does not exist in this registry.
    """
    pass


class UnitParseError(Exception):
    """Raised when a string unit name is not parseable as a valid unit

    Example
    -------

    >>> from unyt import Unit
    >>> Unit('hello')  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.UnitParseError: Could not find unit symbol
    'hello' in the provided symbols.
    """
    pass


class IllDefinedUnitSystem(Exception):
    """Raised when the dimensions of the base units of a unit system are
    inconsistent.

    Example
    -------

    >>> from unyt.unit_systems import UnitSystem
    >>> UnitSystem('atomic', 'nm', 'fs', 'nK', 'rad')\
  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    unyt.exceptions.IllDefinedUnitSystem: Cannot create unit system
    with inconsistent mapping from
    dimensions to units. Received:
    OrderedDict([((length), nm), ((mass), fs), ((time), nK),
                 ((temperature), rad), ((angle), rad),
                 ((current_mks), A), ((luminous_intensity), cd)])
    """
    def __init__(self, units_map):
        self.units_map = units_map

    def __str__(self):
        return ("Cannot create unit system with inconsistent mapping from "
                "dimensions to units. Received:\n%s" % self.units_map)



from unyt.unit_object import define_unit
from unyt.array import unyt_quantity


def test_define_unit():
    define_unit("mph", (1.0, "mile/hr"))
    a = unyt_quantity(2.0, "mph")
    b = unyt_quantity(1.0, "mile")
    c = unyt_quantity(1.0, "hr")
    assert a == 2.0*b/c
    d = unyt_quantity(1000.0, "cm**3")
    define_unit("L", d, prefixable=True)
    e = unyt_quantity(1.0, "mL")
    f = unyt_quantity(1.0, "cm**3")
    assert e == f



"""
Test unit systems.

"""

# -----------------------------------------------------------------------------
# Copyright (c) 2018, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

import pytest

from unyt.exceptions import (
    IllDefinedUnitSystem,
    MissingMKSCurrent,
)
from unyt.unit_object import Unit
from unyt.unit_systems import (
    UnitSystem,
    cgs_unit_system,
    unit_system_registry,
)
from unyt.unit_registry import UnitRegistry
from unyt import dimensions


def test_unit_systems():
    goofy_unit_system = UnitSystem(
        "goofy", "ly", "lbm", "hr", temperature_unit="R",
        angle_unit="arcsec", current_mks_unit="mA",
        luminous_intensity_unit="cd")
    assert goofy_unit_system["temperature"] == Unit("R")
    assert goofy_unit_system[dimensions.solid_angle] == Unit("arcsec**2")
    assert goofy_unit_system["energy"] == Unit("lbm*ly**2/hr**2")
    goofy_unit_system["energy"] = "eV"
    assert goofy_unit_system["energy"] == Unit("eV")
    assert goofy_unit_system["magnetic_field_mks"] == Unit("lbm/(hr**2*mA)")
    assert "goofy" in unit_system_registry


def test_unit_system_id():
    reg1 = UnitRegistry()
    reg2 = UnitRegistry()
    assert reg1.unit_system_id == reg2.unit_system_id
    reg1.modify('g', 2.0)
    assert reg1.unit_system_id != reg2.unit_system_id
    reg1 = UnitRegistry()
    reg1.add('dinosaurs', 12.0, dimensions.length)
    assert reg1.unit_system_id != reg2.unit_system_id
    reg1 = UnitRegistry()
    reg1.remove('g')
    assert reg1.unit_system_id != reg2.unit_system_id
    reg1.add('g', 1.0e-3, dimensions.mass, prefixable=True)
    assert reg1.unit_system_id == reg2.unit_system_id


def test_bad_unit_system():
    with pytest.raises(IllDefinedUnitSystem):
        UnitSystem('atomic', 'nm', 'fs', 'nK', 'rad')


def test_mks_current():
    with pytest.raises(MissingMKSCurrent):
        cgs_unit_system[dimensions.current_mks]
    with pytest.raises(MissingMKSCurrent):
        cgs_unit_system[dimensions.magnetic_field]
    with pytest.raises(MissingMKSCurrent):
        cgs_unit_system[dimensions.current_mks] = 'foo'
    with pytest.raises(MissingMKSCurrent):
        cgs_unit_system[dimensions.magnetic_field] = 'bar'
